import pytest

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.QueryFeatureHGVS import QueryFeatureHGVS
from genomics_data_index.storage.model.QueryFeatureHGVSGN import QueryFeatureHGVSGN
from genomics_data_index.storage.model.QueryFeatureMLST import QueryFeatureMLST
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.db import Sample
from genomics_data_index.storage.service.SampleService import SampleService, FeatureExplodeUnknownError


def count_samples(sample_service, variation_service):
    assert 3 == sample_service.count_samples()


def test_samples_with_variants(sample_service, variation_service):
    samples_with_variants = sample_service.get_samples_with_variants('genome')
    assert {'SampleA', 'SampleB', 'SampleC'} == {sample.name for sample in samples_with_variants}


def test_samples_with_variants_on_sequence(sample_service, variation_service):
    samples_with_variants = sample_service.get_samples_with_variants_on_sequence('reference')
    assert {'SampleA', 'SampleB', 'SampleC'} == {sample.name for sample in samples_with_variants}


def test_samples_associated_with_reference(sample_service: SampleService, variation_service):
    samples_reference = sample_service.get_samples_associated_with_reference('genome')
    assert {'SampleA', 'SampleB', 'SampleC'} == {sample.name for sample in samples_reference}


def test_samples_set_associated_with_reference(database, sample_service, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    expected_sample_set = SampleSet([sampleA.id, sampleB.id, sampleC.id])

    samples_set_reference = sample_service.get_samples_set_associated_with_reference('genome')

    assert isinstance(samples_set_reference, SampleSet)
    assert set(samples_set_reference) == set(expected_sample_set)


def test_find_sample_name_ids(database, sample_service, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    sample_ids_dict = sample_service.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])

    assert {sampleA.name: sampleA.id, sampleB.name: sampleB.id, sampleC.name: sampleC.id} == sample_ids_dict


def test_find_sample_name_ids_select1(database, sample_service_select1, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    sample_ids_dict = sample_service_select1.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])

    assert {sampleA.name: sampleA.id, sampleB.name: sampleB.id, sampleC.name: sampleC.id} == sample_ids_dict


def test_find_sample_name_ids_select2(database, sample_service_select2, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    sample_ids_dict = sample_service_select2.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])

    assert {sampleA.name: sampleA.id, sampleB.name: sampleB.id, sampleC.name: sampleC.id} == sample_ids_dict


def test_count_samples_associated_with_reference(sample_service, variation_service):
    assert 3 == sample_service.count_samples_associated_with_reference('genome')


def test_count_samples_associated_with_reference_empty(sample_service, variation_service):
    assert 0 == sample_service.count_samples_associated_with_reference('no_exist')


def test_samples_associated_with_reference_empty(sample_service, variation_service):
    samples_with_variants = sample_service.get_samples_associated_with_reference('no_exist')
    assert set() == {sample.name for sample in samples_with_variants}


def test_samples_with_variants_empty(sample_service, variation_service):
    samples_with_variants = sample_service.get_samples_with_variants('does_not_exist')
    assert set() == set(samples_with_variants)


def test_samples_which_exists(sample_service, variation_service):
    assert {'SampleA', 'SampleB', 'SampleC'} == set(sample_service.which_exists(['SampleA', 'SampleB',
                                                                                 'SampleC', 'not_exist']))


def test_samples_which_exists_only_one(sample_service, variation_service):
    assert {'SampleA'} == set(sample_service.which_exists(['SampleA']))


def test_samples_which_exists_none(sample_service, variation_service):
    assert [] == sample_service.which_exists(['not_exist'])


def test_samples_which_exists_none2(sample_service, variation_service):
    assert [] == sample_service.which_exists([])


def test_get_samples(sample_service, variation_service):
    assert {'SampleA', 'SampleB', 'SampleC'} == {sample.name for sample in sample_service.get_samples()}


def test_get_samples_empty(sample_service):
    assert [] == sample_service.get_samples()


def test_get_existing_samples_by_names_all(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names(['SampleA', 'SampleB', 'SampleC'])
    assert 3 == len(samples)
    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}


def test_get_existing_samples_by_names_two(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names(['SampleA', 'SampleC'])
    assert 2 == len(samples)
    assert {'SampleA', 'SampleC'} == {s.name for s in samples}


def test_get_existing_samples_by_names_one(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names(['SampleC'])
    assert 1 == len(samples)
    assert {'SampleC'} == {s.name for s in samples}


def test_get_existing_samples_by_names_none(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names([])
    assert 0 == len(samples)
    assert set() == {s.name for s in samples}


def test_get_existing_samples_by_names_sample_not_exist(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names(['SampleA', 'NotFound', 'SampleC'])
    assert 2 == len(samples)
    assert {'SampleA', 'SampleC'} == {s.name for s in samples}


def test_get_existing_samples_by_names_sample_not_exis2t(sample_service, variation_service):
    samples = sample_service.get_existing_samples_by_names(['SampleA', 'NotFound', 'SampleC', 'NotFound2'])
    assert 2 == len(samples)
    assert {'SampleA', 'SampleC'} == {s.name for s in samples}


def test_find_samples_by_ids(sample_service, variation_service):
    sample_name_ids = sample_service.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])
    sample_set = SampleSet(sample_ids=[sample_name_ids['SampleA']])

    samples = sample_service.find_samples_by_ids(sample_set)
    assert {'SampleA'} == {s.name for s in samples}


def test_find_samples_by_ids_2_samples(sample_service, variation_service):
    sample_name_ids = sample_service.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])
    sample_set = SampleSet(sample_ids=[sample_name_ids['SampleA'], sample_name_ids['SampleC']])

    samples = sample_service.find_samples_by_ids(sample_set)
    assert {'SampleA', 'SampleC'} == {s.name for s in samples}


def test_find_samples_by_ids_3_samples(sample_service, variation_service):
    sample_name_ids = sample_service.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])
    sample_set = SampleSet(sample_ids=[sample_name_ids['SampleA'], sample_name_ids['SampleB'],
                                       sample_name_ids['SampleC']])

    samples = sample_service.find_samples_by_ids(sample_set)
    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}


def test_find_samples_by_ids_3_samples_select1(sample_service_select1, variation_service):
    sample_name_ids = sample_service_select1.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])
    sample_set = SampleSet(sample_ids=[sample_name_ids['SampleA'], sample_name_ids['SampleB'],
                                       sample_name_ids['SampleC']])

    samples = sample_service_select1.find_samples_by_ids(sample_set)
    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}


def test_find_samples_by_ids_3_samples_select2(sample_service_select2, variation_service):
    sample_name_ids = sample_service_select2.find_sample_name_ids(['SampleA', 'SampleB', 'SampleC'])
    sample_set = SampleSet(sample_ids=[sample_name_ids['SampleA'], sample_name_ids['SampleB'],
                                       sample_name_ids['SampleC']])

    samples = sample_service_select2.find_samples_by_ids(sample_set)
    assert {'SampleA', 'SampleB', 'SampleC'} == {s.name for s in samples}


def test_find_samples_by_features_variations(database, sample_service, variation_service):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    variant_samples = sample_service.find_samples_by_features([QueryFeatureMutationSPDI('reference:5061:G:A')])

    assert f'reference:5061:G:A' in variant_samples
    assert {'SampleB'} == {s.name for s in variant_samples[f'reference:5061:G:A']}
    assert {sampleB.id} == {s.id for s in variant_samples[f'reference:5061:G:A']}


def test_find_samples_by_features_variations_hgvs(database, sample_service_snpeff_annotations):
    sample_sh14_001 = database.get_session().query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_sh14_014 = database.get_session().query(Sample).filter(Sample.name == 'SH14-014').one()

    # hgvs c (nucleotide)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:c.670dupA')]
    variant_samples = sample_service_snpeff_annotations.find_samples_by_features(features)
    assert 1 == len(variant_samples)
    assert 'hgvs:NC_011083:SEHA_RS04550:c.670dupA' in variant_samples
    assert {'SH14-001', 'SH14-014'} == {s.name for s in variant_samples['hgvs:NC_011083:SEHA_RS04550:c.670dupA']}
    assert {sample_sh14_001.id, sample_sh14_014.id} == {s.id for s in
                                                        variant_samples['hgvs:NC_011083:SEHA_RS04550:c.670dupA']}

    # hgvs p (protein)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:p.Ile224fs')]
    variant_samples = sample_service_snpeff_annotations.find_samples_by_features(features)
    assert 1 == len(variant_samples)
    assert 'hgvs:NC_011083:SEHA_RS04550:p.Ile224fs' in variant_samples
    assert {'SH14-001', 'SH14-014'} == {s.name for s in variant_samples['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs']}
    assert {sample_sh14_001.id, sample_sh14_014.id} == {s.id for s in
                                                        variant_samples['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs']}

    # spdi
    features = [QueryFeatureMutationSPDI('NC_011083:835147:C:CA')]
    variant_samples = sample_service_snpeff_annotations.find_samples_by_features(features)
    assert 1 == len(variant_samples)
    assert 'NC_011083:835147:C:CA' in variant_samples
    assert {'SH14-001', 'SH14-014'} == {s.name for s in variant_samples['NC_011083:835147:C:CA']}
    assert {sample_sh14_001.id, sample_sh14_014.id} == {s.id for s in variant_samples['NC_011083:835147:C:CA']}


def test_find_sample_sets_by_features_variations(database, sample_service, variation_service):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    sample_sets = sample_service.find_sample_sets_by_features([QueryFeatureMutationSPDI('reference:5061:G:A')])

    assert f'reference:5061:G:A' in sample_sets
    assert {sampleB.id} == set(sample_sets[f'reference:5061:G:A'])


def test_find_unknown_sample_sets_by_features_variations_no_index_unknowns(database, sample_service, variation_service):
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:5061:G:A')])
    assert 0 == len(sample_sets)


def test_find_unknown_sample_sets_by_features_variations_with_index_unknowns(database,
                                                                             sample_service,
                                                                             variation_service_index_unknowns):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()

    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:5061:G:A')])

    assert 1 == len(sample_sets)
    assert f'reference:5061:G:A' in sample_sets
    assert {sampleA.id} == set(sample_sets[f'reference:5061:G:A'])


def test_find_unknown_sample_sets_by_features_variations_with_index_unknowns_multiple(database,
                                                                                      sample_service,
                                                                                      variation_service_index_unknowns):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    # Test 3 features, two of which are in same location
    features = [QueryFeatureMutationSPDI('reference:5061:G:A'),
                QueryFeatureMutationSPDI('reference:87:1:A'),
                QueryFeatureMutationSPDI('reference:87:G:T'),
                ]
    sample_sets = sample_service.find_unknown_sample_sets_by_features(features)

    assert 3 == len(sample_sets)
    assert {sampleA.id} == set(sample_sets[f'reference:5061:G:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:1:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:G:T'])

    # Test 3 features, two of which are in same location, one is an insertion
    features = [QueryFeatureMutationSPDI('reference:5061:G:A'),
                QueryFeatureMutationSPDI('reference:87:ATCG:A'),
                QueryFeatureMutationSPDI('reference:87:G:T'),
                ]
    sample_sets = sample_service.find_unknown_sample_sets_by_features(features)

    assert 3 == len(sample_sets)
    assert {sampleA.id} == set(sample_sets[f'reference:5061:G:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:ATCG:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:G:T'])

    # Test 3 features, two of which are in same location, one is a deletion
    features = [QueryFeatureMutationSPDI('reference:5061:G:A'),
                QueryFeatureMutationSPDI('reference:87:T:A'),
                QueryFeatureMutationSPDI('reference:87:G:TCG'),
                ]
    sample_sets = sample_service.find_unknown_sample_sets_by_features(features)

    assert 3 == len(sample_sets)
    assert {sampleA.id} == set(sample_sets[f'reference:5061:G:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:T:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:G:TCG'])

    # Test 4 features, three are complex, one with integer deletion
    features = [QueryFeatureMutationSPDI('reference:5061:G:A'),
                QueryFeatureMutationSPDI('reference:87:TCCG:AAAGG'),
                QueryFeatureMutationSPDI('reference:87:GGGGA:TCG'),
                QueryFeatureMutationSPDI('reference:87:3:TCG'),
                ]
    sample_sets = sample_service.find_unknown_sample_sets_by_features(features)

    assert 4 == len(sample_sets)
    assert {sampleA.id} == set(sample_sets[f'reference:5061:G:A'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:TCCG:AAAGG'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:GGGGA:TCG'])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:3:TCG'])


def test_find_unknown_sample_sets_by_features_hgvs(database,
                                                   sample_service_snpeff_annotations: SampleService):
    sample_service = sample_service_snpeff_annotations

    # Test multiple HGVSGN features, should be no unknowns
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu'),
                QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC')
                ]
    sample_sets = sample_service.find_unknown_sample_sets_by_features(features)

    assert 0 == len(sample_sets)

    # TODO: Expand tests here. I need to add more test data to test cases where I do find unknown sample sets


def test_find_unknown_sample_sets_by_features_variations_different_feature_definitions(database,
                                                                                       sample_service,
                                                                                       variation_service_index_unknowns):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    # Test 3 unknowns
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:87:1:A')])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:87:1:A'])

    # Test 2 unknowns
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:5088:1:G')])
    assert {sampleA.id, sampleC.id} == set(sample_sets[f'reference:5088:1:G'])

    # Test 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:202:1:GCG')])
    assert {sampleA.id} == set(sample_sets[f'reference:202:1:GCG'])

    # Test right before unknown: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:220:1:A')])
    assert 0 == len(sample_sets)

    # Test on unknown: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:221:1:A')])
    assert {sampleA.id} == set(sample_sets[f'reference:221:1:A'])

    # Test right after unknown: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:222:1:A')])
    assert 0 == len(sample_sets)

    # Test right before unknown but 2 bp: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:220:2:A')])
    assert {sampleA.id} == set(sample_sets[f'reference:220:2:A'])

    # Test right before unknown but 3 bp: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:220:3:A')])
    assert {sampleA.id} == set(sample_sets[f'reference:220:3:A'])

    # Test right before unknown but 3 bp as characters: 1 unknown
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:220:ATT:A')])
    assert {sampleA.id} == set(sample_sets[f'reference:220:ATT:A'])

    # Test 3 unknowns 1 bp: before
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:133:1:A')])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:133:1:A'])

    # Test 2 unknowns 1 bp: left edge
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:134:1:A')])
    assert {sampleA.id, sampleC.id} == set(sample_sets[f'reference:134:1:A'])

    # Test 2 unknowns 1 bp: right edge
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:135:1:A')])
    assert {sampleA.id, sampleC.id} == set(sample_sets[f'reference:135:1:A'])

    # Test 2 unknowns 2 bp: left and right edge
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:134:2:A')])
    assert {sampleA.id, sampleC.id} == set(sample_sets[f'reference:134:2:A'])

    # Test 3 unknowns 1 bp: after
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:136:1:A')])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:136:1:A'])

    # Test 3 unknowns 2 bp: before and overlapping left edge
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:133:2:A')])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:133:2:A'])

    # Test 3 unknowns 3 bp: left and right edge and into 3 unknown region
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:134:3:A')])
    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_sets[f'reference:134:3:A'])

    # Test on edge of unknown and indels
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1167:1:A')])
    assert 0 == len(sample_sets)

    # Test on deletion (no unknowns since it's a deletion)
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1168:1:A')])
    assert 0 == len(sample_sets)

    # Test right after deletion (still no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1169:1:A')])
    assert 0 == len(sample_sets)

    # Test deletion 2 bp (no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1167:2:A')])
    assert 0 == len(sample_sets)

    # Test deletion 2 bp (no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1167:AT:A')])
    assert 0 == len(sample_sets)

    # Test deletion 3 bp (no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features(
        [QueryFeatureMutationSPDI('reference:1167:ATT:A')])
    assert 0 == len(sample_sets)

    # Test overlapping deletion but with insertion (no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features(
        [QueryFeatureMutationSPDI('reference:1167:1:AGG')])
    assert 0 == len(sample_sets)

    # Test overlapping deletion but with insertion of 2 bp (no unknown)
    sample_sets = sample_service.find_unknown_sample_sets_by_features([QueryFeatureMutationSPDI('reference:1167:2:AT')])
    assert 0 == len(sample_sets)


def test_find_features_spdi_for_hgvsgn(database, sample_service_snpeff_annotations: SampleService):
    # Test hgvs_gn c (gene name is same as locus id)
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS04550:c.670dupA')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert 'NC_011083:835147:1:CA' == features_spdi[0].id

    # Test hgvs_gn p (gene name is same as locus id)
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS04550:p.Ile224fs')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert 'NC_011083:835147:1:CA' == features_spdi[0].id

    # Test hgvs_gn p, different gene name
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert 'NC_011083:140658:1:A' == features_spdi[0].id

    # Test hgvs_gn c, different gene name
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:c.497C>A')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert 'NC_011083:140658:1:A' == features_spdi[0].id

    # Test hgvs_gn p, complex
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:p.PheGlu195TyrAsp')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert f'NC_011083:3535656:{len("CTCAAAA")}:GTCATAG' == features_spdi[0].id

    # Test hgvs_gn c, complex
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert f'NC_011083:3535656:{len("CTCAAAA")}:GTCATAG' == features_spdi[0].id

    # Test hgvs_gn p, deletion
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS05995:p.Glu140fs')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert f'NC_011083:1125996:{len("CG")}:C' == features_spdi[0].id

    # Test hgvs_gn c, deletion
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS05995:c.418delG')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert f'NC_011083:1125996:{len("CG")}:C' == features_spdi[0].id

    # Test case where nothing if found
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:invalid_gene:c.418delG')
    features_spdi = sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)
    assert 0 == len(features_spdi)

    # Make sure only QueryFeatureHGVSGN is handled
    with pytest.raises(Exception) as execinfo:
        feature = QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS05995:c.418delG')
        sample_service_snpeff_annotations.find_features_spdi_for_hgvsgn(feature)

    assert 'Cannot handle feature=hgvs:NC_011083:SEHA_RS05995:c.418delG. Not of type QueryFeatureHGVSGN' in str(
        execinfo.value)


def test_find_features_spdi_for_hgvsgn_duplicate_gene(database,
                                                      sample_service_snpeff_annotations_fake_duplicate_gene: SampleService):
    # Test cases where there's a duplicate of a gene name and so I get two features from the hgvs_gn identifier

    # Test hgvs_gn p
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu')
    features_spdi = sample_service_snpeff_annotations_fake_duplicate_gene.find_features_spdi_for_hgvsgn(feature)
    assert 2 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert isinstance(features_spdi[1], QueryFeatureMutationSPDI)
    assert {'NC_011083:140658:1:A', 'NC_011083:150000:1:A'} == {f.id for f in features_spdi}

    # Test hgvs_gn c
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:c.497C>A')
    features_spdi = sample_service_snpeff_annotations_fake_duplicate_gene.find_features_spdi_for_hgvsgn(feature)
    assert 2 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert isinstance(features_spdi[1], QueryFeatureMutationSPDI)
    assert {'NC_011083:140658:1:A', 'NC_011083:150000:1:A'} == {f.id for f in features_spdi}

    # Test single HGVS p in same dataset
    feature = QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS01460:p.Thr201Met')
    features_spdi = sample_service_snpeff_annotations_fake_duplicate_gene.find_features_spdi_for_hgvsgn(feature)
    assert 1 == len(features_spdi)
    assert isinstance(features_spdi[0], QueryFeatureMutationSPDI)
    assert 'NC_011083:203200:1:T' == features_spdi[0].id


def test_find_sample_sets_by_features_variations_hgvs(database, sample_service_snpeff_annotations):
    sample_sh14_001 = database.get_session().query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_sh14_014 = database.get_session().query(Sample).filter(Sample.name == 'SH14-014').one()

    # hgvs c (nucleotide)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:c.670dupA')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs:NC_011083:SEHA_RS04550:c.670dupA' in sample_sets
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(sample_sets['hgvs:NC_011083:SEHA_RS04550:c.670dupA'])

    # hgvs p (protein)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:p.Ile224fs')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs:NC_011083:SEHA_RS04550:p.Ile224fs' in sample_sets
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(sample_sets['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs'])

    # spdi
    features = [QueryFeatureMutationSPDI('NC_011083:835147:C:CA')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'NC_011083:835147:C:CA' in sample_sets
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(sample_sets['NC_011083:835147:C:CA'])


def test_find_sample_sets_by_features_variations_hgvs_gn(database, sample_service_snpeff_annotations):
    sample_sh14_001 = database.get_session().query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_sh14_014 = database.get_session().query(Sample).filter(Sample.name == 'SH14-014').one()
    sample_sh10_014 = database.get_session().query(Sample).filter(Sample.name == 'SH10-014').one()

    # hgvs c (nucleotide)
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS04550:c.670dupA')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:SEHA_RS04550:c.670dupA' in sample_sets
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(sample_sets['hgvs_gn:NC_011083:SEHA_RS04550:c.670dupA'])

    # hgvs p (protein)
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:SEHA_RS04550:p.Ile224fs')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:SEHA_RS04550:p.Ile224fs' in sample_sets
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(sample_sets['hgvs_gn:NC_011083:SEHA_RS04550:p.Ile224fs'])

    # hgvs c with gene name different
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC' in sample_sets
    assert {sample_sh10_014.id, sample_sh14_001.id} == set(
        sample_sets['hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC'])

    # hgvs p with gene name different
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:p.PheGlu195TyrAsp')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:oadA:p.PheGlu195TyrAsp' in sample_sets
    assert {sample_sh10_014.id, sample_sh14_001.id} == set(sample_sets['hgvs_gn:NC_011083:oadA:p.PheGlu195TyrAsp'])

    # hgvs c with gene name different 2
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:c.497C>A')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:murF:c.497C>A' in sample_sets
    assert {sample_sh10_014.id, sample_sh14_014.id, sample_sh14_001.id} == set(
        sample_sets['hgvs_gn:NC_011083:murF:c.497C>A'])

    # hgvs p with gene name different 2
    features = [QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu')]
    sample_sets = sample_service_snpeff_annotations.find_sample_sets_by_features(features)
    assert 1 == len(sample_sets)
    assert 'hgvs_gn:NC_011083:murF:p.Ala166Glu' in sample_sets
    assert {sample_sh10_014.id, sample_sh14_014.id, sample_sh14_001.id} == set(
        sample_sets['hgvs_gn:NC_011083:murF:p.Ala166Glu'])


def test_find_sample_sets_by_features_variations_two_samples(database, sample_service, variation_service):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    variant_samples = sample_service.find_sample_sets_by_features([QueryFeatureMutationSPDI('reference:4975:T:C')])

    assert {'reference:4975:T:C'} == set(variant_samples.keys())
    assert {sampleC.id, sampleB.id} == set(variant_samples[f'reference:4975:T:C'])
    assert 2 == len(variant_samples[f'reference:4975:T:C'])


def test_find_samples_by_features_variations_numeric_deletion(database, sample_service, variation_service):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    variant_samples = sample_service.find_samples_by_features([QueryFeatureMutationSPDI('reference:5061:1:A')])

    assert f'reference:5061:1:A' in variant_samples
    assert {'SampleB'} == {s.name for s in variant_samples[f'reference:5061:1:A']}
    assert {sampleB.id} == {s.id for s in variant_samples[f'reference:5061:1:A']}


def test_count_samples_by_variation_features_single_feature(sample_service, variation_service):
    features = [QueryFeatureMutationSPDI('reference:5061:G:A')]

    variant_counts = sample_service.count_samples_by_features(features)

    assert 1 == len(variant_counts)
    assert 1 == variant_counts[f'reference:5061:G:A']


def test_count_samples_by_variation_features_multiple_features(sample_service, variation_service):
    features = [QueryFeatureMutationSPDI('reference:5061:G:A'), QueryFeatureMutationSPDI('reference:3063:A:ATGCAGC')]

    variant_counts = sample_service.count_samples_by_features(features)

    assert 2 == len(variant_counts)
    assert 1 == variant_counts[f'reference:5061:G:A']
    assert 2 == variant_counts[f'reference:3063:A:ATGCAGC']


def test_count_samples_by_variation_features_multiple_features_numeric_deletion(sample_service, variation_service):
    features = [QueryFeatureMutationSPDI('reference:5061:1:A'), QueryFeatureMutationSPDI('reference:3063:1:ATGCAGC')]

    variant_counts = sample_service.count_samples_by_features(features)

    assert 2 == len(variant_counts)
    assert 1 == variant_counts[f'reference:5061:1:A']
    assert 2 == variant_counts[f'reference:3063:1:ATGCAGC']


def test_count_samples_by_features_variations_hgvs_multiple_features(database, sample_service_snpeff_annotations):
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:p.Ile224fs'),
                QueryFeatureHGVS.create_from_id('hgvs:NC_011083:n.882634G>A')]
    variant_counts = sample_service_snpeff_annotations.count_samples_by_features(features)
    assert 2 == len(variant_counts)
    assert 1 == variant_counts['hgvs:NC_011083:n.882634G>A']
    assert 2 == variant_counts['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs']


def test_get_samples_with_mlst_alleles(sample_service, mlst_service_loaded):
    samples = sample_service.get_samples_with_mlst_alleles('lmonocytogenes')

    assert {'CFSAN002349', 'CFSAN023463'} == {sample.name for sample in samples}


def test_get_samples_with_mlst_alleles2(sample_service, mlst_service_loaded):
    samples = sample_service.get_samples_with_mlst_alleles('ecoli')

    assert {'2014C-3598', '2014C-3599'} == {sample.name for sample in samples}


def test_find_samples_by_features_mlst(database, sample_service, mlst_service_loaded):
    sample1 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample2 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN023463').one()

    mlst_samples = sample_service.find_samples_by_features([QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1')])

    assert {'mlst:lmonocytogenes:abcZ:1'} == set(mlst_samples.keys())
    assert {'CFSAN002349', 'CFSAN023463'} == {s.name for s in mlst_samples['mlst:lmonocytogenes:abcZ:1']}
    assert {sample1.id, sample2.id} == {s.id for s in mlst_samples['mlst:lmonocytogenes:abcZ:1']}


def test_find_sample_sets_by_features_mlst(database, sample_service: SampleService, mlst_service_loaded):
    sample1 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample2 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN023463').one()

    mlst_sample_sets = sample_service.find_sample_sets_by_features([QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1')])

    assert {'mlst:lmonocytogenes:abcZ:1'} == set(mlst_sample_sets.keys())
    assert {sample1.id, sample2.id} == set(mlst_sample_sets['mlst:lmonocytogenes:abcZ:1'])


def test_find_samples_by_features_mlst_two(database, sample_service, mlst_service_loaded):
    sample1 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample2 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN023463').one()
    sample3 = database.get_session().query(Sample).filter(Sample.name == '2014C-3599').one()
    sample4 = database.get_session().query(Sample).filter(Sample.name == '2014C-3598').one()

    mlst_samples = sample_service.find_samples_by_features([QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1'),
                                                            QueryFeatureMLST('mlst:ecoli:adk:100')])

    assert {'mlst:lmonocytogenes:abcZ:1', 'mlst:ecoli:adk:100'} == set(mlst_samples.keys())

    assert {'CFSAN002349', 'CFSAN023463'} == {s.name for s in mlst_samples['mlst:lmonocytogenes:abcZ:1']}
    assert {sample1.id, sample2.id} == {s.id for s in mlst_samples['mlst:lmonocytogenes:abcZ:1']}

    assert {'2014C-3599', '2014C-3598'} == {s.name for s in mlst_samples['mlst:ecoli:adk:100']}
    assert {sample3.id, sample4.id} == {s.id for s in mlst_samples['mlst:ecoli:adk:100']}


def test_find_sample_sets_by_features_mlst_two(database, sample_service, mlst_service_loaded):
    sample1 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN002349').one()
    sample2 = database.get_session().query(Sample).filter(Sample.name == 'CFSAN023463').one()
    sample3 = database.get_session().query(Sample).filter(Sample.name == '2014C-3599').one()
    sample4 = database.get_session().query(Sample).filter(Sample.name == '2014C-3598').one()

    mlst_sample_sets = sample_service.find_sample_sets_by_features([QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1'),
                                                                    QueryFeatureMLST('mlst:ecoli:adk:100')])

    assert {'mlst:lmonocytogenes:abcZ:1', 'mlst:ecoli:adk:100'} == set(mlst_sample_sets.keys())

    assert {sample1.id, sample2.id} == set(mlst_sample_sets['mlst:lmonocytogenes:abcZ:1'])
    assert {sample3.id, sample4.id} == set(mlst_sample_sets['mlst:ecoli:adk:100'])


def test_count_samples_by_mlst_features_single_feature(sample_service, mlst_service_loaded):
    features = [QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1')]

    mlst_counts = sample_service.count_samples_by_features(features)

    assert 1 == len(mlst_counts)
    assert 2 == mlst_counts['mlst:lmonocytogenes:abcZ:1']


def test_count_samples_by_mlst_features_multiple_features(sample_service, mlst_service_loaded):
    features = [QueryFeatureMLST('mlst:lmonocytogenes:abcZ:1'), QueryFeatureMLST('mlst:ecoli:adk:100')]

    mlst_counts = sample_service.count_samples_by_features(features)

    assert 2 == len(mlst_counts)
    assert 2 == mlst_counts['mlst:lmonocytogenes:abcZ:1']
    assert 2 == mlst_counts['mlst:ecoli:adk:100']


def test_create_dataframe_from_sample_set(database, sample_service: SampleService, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    # Only present
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet([sampleA.id, sampleB.id, sampleC.id]),
                                                         absent_set=SampleSet.create_empty(),
                                                         unknown_set=SampleSet.create_empty(),
                                                         queries_expression='')
    assert len(df) == 3
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleB', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleB.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Present', 'Present', 'Present'] == df['Status'].tolist()

    # Only absent
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet.create_empty(),
                                                         absent_set=SampleSet([sampleA.id, sampleB.id, sampleC.id]),
                                                         unknown_set=SampleSet.create_empty(),
                                                         queries_expression='')
    assert len(df) == 3
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleB', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleB.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Absent', 'Absent', 'Absent'] == df['Status'].tolist()

    # Only unknown
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet.create_empty(),
                                                         absent_set=SampleSet.create_empty(),
                                                         unknown_set=SampleSet([sampleA.id, sampleB.id, sampleC.id]),
                                                         queries_expression='')
    assert len(df) == 3
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleB', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleB.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Unknown', 'Unknown', 'Unknown'] == df['Status'].tolist()

    # Mix of each
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet([sampleA.id]),
                                                         absent_set=SampleSet([sampleB.id]),
                                                         unknown_set=SampleSet([sampleC.id]),
                                                         queries_expression='')
    assert len(df) == 3
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleB', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleB.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Present', 'Absent', 'Unknown'] == df['Status'].tolist()


def test_create_dataframe_from_sample_set_subset_samples(database, sample_service: SampleService, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    # Both present
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet([sampleA.id, sampleC.id]),
                                                         absent_set=SampleSet.create_empty(),
                                                         unknown_set=SampleSet.create_empty(),
                                                         queries_expression='')
    assert len(df) == 2
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Present', 'Present'] == df['Status'].tolist()

    # Mix
    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet([sampleA.id]),
                                                         absent_set=SampleSet.create_empty(),
                                                         unknown_set=SampleSet([sampleC.id]),
                                                         queries_expression='')
    assert len(df) == 2
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()
    df = df.sort_values(['Sample Name'])
    assert ['SampleA', 'SampleC'] == df['Sample Name'].tolist()
    assert [sampleA.id, sampleC.id] == df['Sample ID'].tolist()
    assert ['Present', 'Unknown'] == df['Status'].tolist()


def test_create_dataframe_from_sample_set_empty(sample_service: SampleService, variation_service):
    sample_set = SampleSet.create_empty()
    queries_expression = ''

    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet.create_empty(),
                                                         absent_set=SampleSet.create_empty(),
                                                         unknown_set=SampleSet.create_empty(),
                                                         queries_expression='')
    assert len(df) == 0
    assert ['Query', 'Sample Name', 'Sample ID', 'Status'] == df.columns.tolist()


def test_create_dataframe_from_sample_set_with_query_expression(database, sample_service: SampleService,
                                                                variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    df = sample_service.create_dataframe_from_sample_set(present_set=SampleSet([sampleA.id]),
                                                         absent_set=SampleSet([sampleB.id]),
                                                         unknown_set=SampleSet.create_empty(),
                                                         queries_expression='lmonocytogenes:abc:1')
    assert len(df) == 2
    assert {'lmonocytogenes:abc:1'} == set(df['Query'].tolist())


def test_get_all_sample_ids(database, sample_service, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    assert {sampleA.id, sampleB.id, sampleC.id} == set(sample_service.get_all_sample_ids())


def test_get_variants_samples_by_variation_features_only_spdi(database, sample_service, variation_service):
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()

    feature_id_nucleotide_samples = sample_service.get_variants_samples_by_variation_features(
        [QueryFeatureMutationSPDI('reference:5061:G:A')])

    assert 1 == len(feature_id_nucleotide_samples)
    assert f'reference:5061:G:A' in feature_id_nucleotide_samples
    assert {sampleB.id} == set(feature_id_nucleotide_samples[f'reference:5061:G:A'].sample_ids)

    # Test case of two features which are identical but different names
    feature_id_nucleotide_samples = sample_service.get_variants_samples_by_variation_features(
        [QueryFeatureMutationSPDI('reference:5061:G:A'),
         QueryFeatureMutationSPDI('reference:5061:1:A')])

    assert 2 == len(feature_id_nucleotide_samples)
    assert {sampleB.id} == set(feature_id_nucleotide_samples[f'reference:5061:G:A'].sample_ids)
    assert {sampleB.id} == set(feature_id_nucleotide_samples[f'reference:5061:1:A'].sample_ids)


def test_get_variants_samples_by_variation_features_only_hgvs_c(database, sample_service_snpeff_annotations):
    sample_sh10_014 = database.get_session().query(Sample).filter(Sample.name == 'SH10-014').one()
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:n.882634G>A')]

    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)

    assert 1 == len(feature_id_nucleotide_samples)
    assert 'hgvs:NC_011083:n.882634G>A' in feature_id_nucleotide_samples
    assert {sample_sh10_014.id} == set(feature_id_nucleotide_samples['hgvs:NC_011083:n.882634G>A'].sample_ids)


def test_get_variants_samples_by_variation_features_only_hgvs_cp_spdi(database, sample_service_snpeff_annotations):
    sample_sh14_001 = database.get_session().query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_sh14_014 = database.get_session().query(Sample).filter(Sample.name == 'SH14-014').one()

    # hgvs c (nucleotide)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:c.670dupA')]
    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)
    assert 1 == len(feature_id_nucleotide_samples)
    assert 'hgvs:NC_011083:SEHA_RS04550:c.670dupA' in feature_id_nucleotide_samples
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['hgvs:NC_011083:SEHA_RS04550:c.670dupA'].sample_ids)

    # hgvs p (protein)
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:p.Ile224fs')]
    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)
    assert 1 == len(feature_id_nucleotide_samples)
    assert 'hgvs:NC_011083:SEHA_RS04550:p.Ile224fs' in feature_id_nucleotide_samples
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs'].sample_ids)

    # spdi
    features = [QueryFeatureMutationSPDI('NC_011083:835147:C:CA')]
    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)
    assert 1 == len(feature_id_nucleotide_samples)
    assert 'NC_011083:835147:C:CA' in feature_id_nucleotide_samples
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['NC_011083:835147:C:CA'].sample_ids)


def test_get_variants_samples_by_variation_features_both_hgvs_spdi(database, sample_service_snpeff_annotations):
    sample_sh14_001 = database.get_session().query(Sample).filter(Sample.name == 'SH14-001').one()
    sample_sh14_014 = database.get_session().query(Sample).filter(Sample.name == 'SH14-014').one()
    sample_sh10_014 = database.get_session().query(Sample).filter(Sample.name == 'SH10-014').one()

    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:c.670dupA'),
                QueryFeatureHGVS.create_from_id('hgvs:NC_011083:n.882634G>A'),
                QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS04550:p.Ile224fs'),
                QueryFeatureMutationSPDI('NC_011083:835147:C:CA')]
    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)

    assert 4 == len(feature_id_nucleotide_samples)
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['hgvs:NC_011083:SEHA_RS04550:c.670dupA'].sample_ids)
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['hgvs:NC_011083:SEHA_RS04550:p.Ile224fs'].sample_ids)
    assert {sample_sh10_014.id} == set(feature_id_nucleotide_samples['hgvs:NC_011083:n.882634G>A'].sample_ids)
    assert {sample_sh14_001.id, sample_sh14_014.id} == set(
        feature_id_nucleotide_samples['NC_011083:835147:C:CA'].sample_ids)


def test_get_variants_samples_by_variation_features_no_matches(database, sample_service_snpeff_annotations):
    features = [QueryFeatureHGVS.create_from_id('hgvs:NC_011083:n.unknown')]

    feature_id_nucleotide_samples = sample_service_snpeff_annotations.get_variants_samples_by_variation_features(
        features)

    assert 0 == len(feature_id_nucleotide_samples)


def test_feature_explode_unknown(sample_service_snpeff_annotations):
    sample_service = sample_service_snpeff_annotations

    # MLST
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMLST('mlst:ecoli:abc:1'))
    assert 1 == len(unknowns)
    assert 'mlst:ecoli:abc:?' == unknowns[0].id
    assert 'ecoli:abc:?' == unknowns[0].id_no_prefix

    # MLST 2
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMLST('mlst:ecoli:abc:15'))
    assert 1 == len(unknowns)
    assert 'mlst:ecoli:abc:?' == unknowns[0].id
    assert 'ecoli:abc:?' == unknowns[0].id_no_prefix

    # SPDI and SNV, does not exist
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMutationSPDI('NC_011083:1:G:A'))
    assert 1 == len(unknowns)
    assert 'NC_011083:1:G:?' == unknowns[0].id

    # SPDI and SNV, does not exist, deletion is number
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMutationSPDI('NC_011083:1:1:A'))
    assert 1 == len(unknowns)
    assert 'NC_011083:1:1:?' == unknowns[0].id

    # SPDI and SNV does exist
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMutationSPDI('NC_011083:4384633:G:A'))
    assert 1 == len(unknowns)
    assert 'NC_011083:4384633:G:?' == unknowns[0].id

    # SPDI and deletion
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMutationSPDI('NC_011083:3425558:AGCC:A'))
    assert 4 == len(unknowns)
    assert ['NC_011083:3425558:A:?', 'NC_011083:3425559:G:?',
            'NC_011083:3425560:C:?', 'NC_011083:3425561:C:?'] == [u.id for u in unknowns]

    # SPDI and insertion
    unknowns = sample_service.feature_explode_unknown(QueryFeatureMutationSPDI('NC_011083:1944163:T:TGGC'))
    assert 1 == len(unknowns)
    assert 'NC_011083:1944163:T:?' == unknowns[0].id

    # HGVS and SNV, nucleotide
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS21795:c.798G>A'))
    assert 1 == len(unknowns)
    assert 'NC_011083:4384633:1:?' == unknowns[0].id

    # HGVS and SNV, protein
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS21795:p.Lys266Lys'))
    assert 1 == len(unknowns)
    assert 'NC_011083:4384633:1:?' == unknowns[0].id

    # HGVS and deletion, nucleotide
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS17245:c.123_125delGGC'))
    assert 4 == len(unknowns)
    assert ['NC_011083:3425558:1:?', 'NC_011083:3425559:1:?',
            'NC_011083:3425560:1:?', 'NC_011083:3425561:1:?'] == [u.id for u in unknowns]

    # HGVS and deletion, protein
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS17245:p.Ala42del'))
    assert 4 == len(unknowns)
    assert ['NC_011083:3425558:1:?', 'NC_011083:3425559:1:?',
            'NC_011083:3425560:1:?', 'NC_011083:3425561:1:?'] == [u.id for u in unknowns]

    # HGVS and insertion, nucleotide
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS10100:c.1293_1295dupGGC'))
    assert 1 == len(unknowns)
    assert 'NC_011083:1944163:1:?' == unknowns[0].id

    # HGVS and complex/other, protein
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS17780:p.ArgAla374HisThr'))
    assert 5 == len(unknowns)
    assert ['NC_011083:3535121:1:?', 'NC_011083:3535122:1:?',
            'NC_011083:3535123:1:?', 'NC_011083:3535124:1:?',
            'NC_011083:3535125:1:?'] == [u.id for u in unknowns]

    # Make sure exception is thrown when it cannot find HGVS
    with pytest.raises(FeatureExplodeUnknownError) as execinfo:
        sample_service.feature_explode_unknown(
            QueryFeatureHGVS.create_from_id('hgvs:NC_011083:SEHA_RS21795:p.Lys266Ala'))

    assert 'feature=hgvs:NC_011083:SEHA_RS21795:p.Lys266Ala is of type HGVS' \
           ' but the corresponding SPDI feature does not exist in the database' in str(execinfo.value)

    # HGVSGN.c and SNV
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:c.497C>A'))
    assert 1 == len(unknowns)
    assert ['NC_011083:140658:1:?'] == [u.id for u in unknowns]

    # HGVSGN.p and SNV
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu'))
    assert 1 == len(unknowns)
    assert ['NC_011083:140658:1:?'] == [u.id for u in unknowns]

    # HGVSGN.c and complex
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:c.582_588delTTTTGAGinsCTATGAC'))
    assert 7 == len('TTTTGAG')
    assert 7 == len(unknowns)
    assert [f'NC_011083:{3535656 + offset}:1:?' for offset in range(7)] == [u.id for u in unknowns]

    # HGVSGN.p and complex
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:oadA:p.PheGlu195TyrAsp'))
    assert 7 == len('TTTTGAG')
    assert 7 == len(unknowns)
    assert [f'NC_011083:{3535656 + offset}:1:?' for offset in range(7)] == [u.id for u in unknowns]

    # Make sure exception is thrown when it cannot find HGVSGN identifier
    with pytest.raises(FeatureExplodeUnknownError) as execinfo:
        sample_service.feature_explode_unknown(
            QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:invalid_gene:p.PheGlu195TyrAsp'))

    assert 'feature=hgvs_gn:NC_011083:invalid_gene:p.PheGlu195TyrAsp is of type HGVSGN' \
           ' but the corresponding SPDI feature does not exist in the database' in str(execinfo.value)


def test_feature_explode_unknown_duplicate_gene_name(sample_service_snpeff_annotations_fake_duplicate_gene):
    sample_service = sample_service_snpeff_annotations_fake_duplicate_gene

    # HGVSGN.c and SNV
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:c.497C>A'))
    assert 2 == len(unknowns)
    assert ['NC_011083:140658:1:?', 'NC_011083:150000:1:?'] == [u.id for u in unknowns]

    # HGVSGN.p and SNV
    unknowns = sample_service.feature_explode_unknown(
        QueryFeatureHGVSGN.create_from_id('hgvs_gn:NC_011083:murF:p.Ala166Glu'))
    assert 2 == len(unknowns)
    assert ['NC_011083:140658:1:?', 'NC_011083:150000:1:?'] == [u.id for u in unknowns]


def test_get_sample_set_by_names(database, sample_service: SampleService, variation_service):
    sampleA = database.get_session().query(Sample).filter(Sample.name == 'SampleA').one()
    sampleB = database.get_session().query(Sample).filter(Sample.name == 'SampleB').one()
    sampleC = database.get_session().query(Sample).filter(Sample.name == 'SampleC').one()

    # Test getting sample sets
    assert {sampleA.id} == set(sample_service.get_sample_set_by_names(['SampleA']))
    assert {sampleB.id} == set(sample_service.get_sample_set_by_names(['SampleB']))
    assert {sampleA.id, sampleB.id} == set(sample_service.get_sample_set_by_names(['SampleA', 'SampleB']))
    assert {sampleA.id, sampleB.id} == set(sample_service.get_sample_set_by_names({'SampleA', 'SampleB'}))
    assert {sampleA.id, sampleB.id, sampleC.id} == set(
        sample_service.get_sample_set_by_names(['SampleA', 'SampleB', 'SampleC']))

    # Test empty sample set
    sample_set = sample_service.get_sample_set_by_names([])
    assert isinstance(sample_set, SampleSet)
    assert 0 == len(sample_set)

    # Test case of trying to get ids for samples that don't exist in database.
    with pytest.raises(Exception) as execinfo:
        sample_service.get_sample_set_by_names(['SampleA', 'Sample_invalid'])
    assert 'Did not find an equal number of sample names and ids' in str(execinfo.value)

    # Test case of ignoring not found samples
    assert {sampleA.id} == set(sample_service.get_sample_set_by_names(['SampleA', 'Sample_invalid'],
                                                                      ignore_not_found=True))



from pathlib import Path
from tempfile import TemporaryDirectory
from typing import List, Union

import pandas as pd
import vcf
from pybedtools import BedTool

from genomics_data_index.pipelines.SnakemakePipelineExecutor import SnakemakePipelineExecutor
from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile
from genomics_data_index.test.integration import reference_file, reference_file_5000_snpeff_2
from genomics_data_index.test.integration.pipelines import assemblies_samples, assemblies_reference, expected_mutations
from genomics_data_index.test.integration.pipelines import empty_bed_file, expected_beds
from genomics_data_index.test.integration.pipelines import snpeff_input_sampleA, snpeff_reference_genome
from genomics_data_index.test.integration.pipelines import snpeff_reads_paired, snpeff_reads_single


def vcf_to_mutations_list(vcf_file: Path) -> List[str]:
    reader = vcf.Reader(filename=str(vcf_file))
    df = pd.DataFrame([vars(r) for r in reader])
    df['ALT'] = df['ALT'].apply(lambda x: x[0])
    mutations = df.apply(lambda x: f'{x["CHROM"]}:{x["POS"]}:{x["REF"]}:{x["ALT"]}', axis='columns')

    return mutations.tolist()


def read_expected_mutations(file: Path) -> List[str]:
    with open(file, 'r') as fh:
        mutations = [l.strip() for l in fh.readlines()]
        return mutations


def assert_vcf(vcf_file: Path, expected_mutations_file: Path):
    sample_expected_mutations = read_expected_mutations(expected_mutations_file)
    actual_mutations = vcf_to_mutations_list(vcf_file)
    assert len(actual_mutations) == len(sample_expected_mutations)
    assert actual_mutations == sample_expected_mutations


def assert_masks_equal(actual_mask_file: Path, expected_bed: Union[Path, BedTool]):
    actual_mask = BedTool(str(actual_mask_file))

    if isinstance(expected_bed, Path):
        expected_bed = BedTool(str(expected_bed))

    assert expected_bed == actual_mask, f'expected=\n{expected_bed} != actual=\n{actual_mask}'


def test_create_fofn_file_single_sample():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_file = tmp_dir / 'assemblies' / 'variant' / 'SampleA.vcf.gz'
        actual_mask_file = tmp_dir / 'assemblies' / 'mask' / 'SampleA.bed.gz'
        unexpected_mlst_file = tmp_dir / 'mlst.tsv'
        input_samples = [assemblies_samples['SampleA']]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()
        assert actual_mutations_file.exists()
        assert actual_mask_file.exists()
        assert not unexpected_mlst_file.exists()  # Make sure MLST file does not exist in this case

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 1 == len(fofn_df)
        assert ['SampleA'] == fofn_df['Sample'].tolist()
        actual_mutations_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])

        assert actual_mutations_file == actual_mutations_file_from_df
        assert actual_mask_file == actual_mask_file_from_df

        assert_vcf(actual_mutations_file, expected_mutations['SampleA'])
        assert_masks_equal(actual_mask_file, empty_bed_file)


def test_create_fofn_file_multiple_samples():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        samples = ['SampleA', 'SampleB', 'SampleC']

        input_samples = [assemblies_samples[s] for s in samples]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 3 == len(fofn_df)
        assert ['SampleA', 'SampleB', 'SampleC'] == fofn_df['Sample'].tolist()

        actual_mutations_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_A, expected_mutations['SampleA'])
        assert_masks_equal(actual_mask_file_A, empty_bed_file)

        actual_mutations_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB']['VCF'].tolist()[0])
        actual_mask_file_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_B, expected_mutations['SampleB'])
        assert_masks_equal(actual_mask_file_B, empty_bed_file)

        actual_mutations_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC']['VCF'].tolist()[0])
        actual_mask_file_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_C, expected_mutations['SampleC'])
        assert_masks_equal(actual_mask_file_C, empty_bed_file)


def test_create_fofn_file_multiple_samples_with_invalid_charcaters_in_name():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        samples_original = ['SampleA', 'SampleB', 'SampleC']
        input_samples = [assemblies_samples[s] for s in samples_original]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)

        # Modify sample names so they contain extra characters that can't be used for a file name
        # (in this case, a slash '/')
        sample_files['Sample'] = sample_files['Sample'] + '/extra|1'
        expected_mutations_local = {s + '/extra|1': expected_mutations[s] for s in expected_mutations}

        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 3 == len(fofn_df)
        assert ['SampleA/extra|1', 'SampleB/extra|1', 'SampleC/extra|1'] == fofn_df['Sample'].tolist()

        actual_mutations_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA/extra|1']['VCF'].tolist()[0])
        actual_mask_file_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA/extra|1']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_A, expected_mutations_local['SampleA/extra|1'])
        assert_masks_equal(actual_mask_file_A, empty_bed_file)

        actual_mutations_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB/extra|1']['VCF'].tolist()[0])
        actual_mask_file_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB/extra|1']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_B, expected_mutations_local['SampleB/extra|1'])
        assert_masks_equal(actual_mask_file_B, empty_bed_file)

        actual_mutations_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC/extra|1']['VCF'].tolist()[0])
        actual_mask_file_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC/extra|1']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_C, expected_mutations_local['SampleC/extra|1'])
        assert_masks_equal(actual_mask_file_C, empty_bed_file)


def test_create_fofn_file_multiple_samples_batching():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        samples = ['SampleA', 'SampleB', 'SampleC']

        input_samples = [assemblies_samples[s] for s in samples]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False, snakemake_input_batch_size=2)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 3 == len(fofn_df)
        assert ['SampleA', 'SampleB', 'SampleC'] == fofn_df['Sample'].tolist()

        actual_mutations_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_A = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_A, expected_mutations['SampleA'])
        assert_masks_equal(actual_mask_file_A, empty_bed_file)

        actual_mutations_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB']['VCF'].tolist()[0])
        actual_mask_file_B = Path(fofn_df[fofn_df['Sample'] == 'SampleB']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_B, expected_mutations['SampleB'])
        assert_masks_equal(actual_mask_file_B, empty_bed_file)

        actual_mutations_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC']['VCF'].tolist()[0])
        actual_mask_file_C = Path(fofn_df[fofn_df['Sample'] == 'SampleC']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_C, expected_mutations['SampleC'])
        assert_masks_equal(actual_mask_file_C, empty_bed_file)


def test_create_fofn_file_multiple_samples_with_ns_deletions():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        samples = ['SampleD', 'SampleE', 'SampleF', 'SampleH']

        input_samples = [assemblies_samples[s] for s in samples]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 4 == len(fofn_df)
        assert ['SampleD', 'SampleE', 'SampleF', 'SampleH'] == fofn_df['Sample'].tolist()

        actual_mutations_D = Path(fofn_df[fofn_df['Sample'] == 'SampleD']['VCF'].tolist()[0])
        actual_mask_D = Path(fofn_df[fofn_df['Sample'] == 'SampleD']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_D, expected_mutations['SampleD'])
        assert_masks_equal(actual_mask_D, expected_beds['SampleD'])

        actual_mutations_E = Path(fofn_df[fofn_df['Sample'] == 'SampleE']['VCF'].tolist()[0])
        actual_mask_E = Path(fofn_df[fofn_df['Sample'] == 'SampleE']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_E, expected_mutations['SampleE'])
        assert_masks_equal(actual_mask_E, expected_beds['SampleE'])

        actual_mutations_F = Path(fofn_df[fofn_df['Sample'] == 'SampleF']['VCF'].tolist()[0])
        actual_mask_F = Path(fofn_df[fofn_df['Sample'] == 'SampleF']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_F, expected_mutations['SampleF'])
        assert_masks_equal(actual_mask_F, expected_beds['SampleF'])

        actual_mutations_H = Path(fofn_df[fofn_df['Sample'] == 'SampleH']['VCF'].tolist()[0])
        actual_mask_H = Path(fofn_df[fofn_df['Sample'] == 'SampleH']['Mask File'].tolist()[0])
        assert_vcf(actual_mutations_H, expected_mutations['SampleH'])
        assert_masks_equal(actual_mask_H, expected_beds['SampleH'])


def test_create_fofn_file_multiple_samples_multiple_cores_and_use_conda():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        samples = ['SampleA', 'SampleB', 'SampleC', 'SampleD', 'SampleE', 'SampleF']

        input_samples = [assemblies_samples[s] for s in samples]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=True,
                                                      include_mlst=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=2)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 6 == len(fofn_df)
        assert ['SampleA', 'SampleB', 'SampleC', 'SampleD', 'SampleE', 'SampleF'] == fofn_df['Sample'].tolist()


def test_create_fofn_file_single_sketch_mlst():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_file = tmp_dir / 'assemblies' / 'variant' / 'SampleA.vcf.gz'
        actual_mask_file = tmp_dir / 'assemblies' / 'mask' / 'SampleA.bed.gz'
        actual_sketch_file = tmp_dir / 'assemblies' / 'sketch' / 'SampleA.sig.gz'
        actual_mlst_file = tmp_dir / 'mlst.tsv'
        input_samples = [assemblies_samples['SampleA']]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=True)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')
        mlst_file = results.get_file('mlst')

        assert input_fofn.exists()
        assert actual_mutations_file.exists()
        assert actual_mask_file.exists()
        assert actual_sketch_file.exists()
        assert actual_mlst_file.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File', 'Sketch File'] == fofn_df.columns.tolist()

        assert 1 == len(fofn_df)
        assert ['SampleA'] == fofn_df['Sample'].tolist()
        actual_mutations_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])
        actual_sketch_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Sketch File'].tolist()[0])

        assert actual_mutations_file == actual_mutations_file_from_df
        assert actual_mask_file == actual_mask_file_from_df
        assert actual_sketch_file == actual_sketch_file_from_df
        assert actual_mlst_file == mlst_file


def test_create_fofn_file_single_no_sketch_with_mlst():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_file = tmp_dir / 'assemblies' / 'variant' / 'SampleA.vcf.gz'
        actual_mask_file = tmp_dir / 'assemblies' / 'mask' / 'SampleA.bed.gz'
        unexpected_sketch_file = tmp_dir / 'assemblies' / 'sketch' / 'SampleA.sig.gz'
        actual_mlst_file = tmp_dir / 'mlst.tsv'
        input_samples = [assemblies_samples['SampleA']]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=True, include_kmer=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=assemblies_reference,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')
        mlst_file = results.get_file('mlst')

        assert input_fofn.exists()
        assert actual_mutations_file.exists()
        assert actual_mask_file.exists()
        assert not unexpected_sketch_file.exists()
        assert actual_mlst_file.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        print(fofn_df)
        assert ['Sample', 'VCF', 'Mask File'] == fofn_df.columns.tolist()

        assert 1 == len(fofn_df)
        assert ['SampleA'] == fofn_df['Sample'].tolist()
        actual_mutations_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])

        assert actual_mutations_file == actual_mutations_file_from_df
        assert actual_mask_file == actual_mask_file_from_df
        assert actual_mlst_file == mlst_file


def test_create_fofn_file_snpeff_no_conda():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_snpeff_file = tmp_dir / 'assemblies' / 'variant-snpeff' / 'SampleA.vcf.gz'
        actual_snpeff_config = tmp_dir / 'snpeff_db' / 'snpEff.config'
        actual_mask_file = tmp_dir / 'assemblies' / 'mask' / 'SampleA.bed.gz'
        input_samples = [snpeff_input_sampleA]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=False,
                                                      include_mlst=False, include_kmer=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=snpeff_reference_genome,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()
        assert actual_snpeff_config.exists()
        assert actual_mutations_snpeff_file.exists()
        assert actual_mask_file.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        assert ['Sample', 'VCF', 'Mask File'] == fofn_df.columns.tolist()

        assert 1 == len(fofn_df)
        assert ['SampleA'] == fofn_df['Sample'].tolist()
        actual_mutations_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])

        assert actual_mutations_snpeff_file == actual_mutations_file_from_df
        assert actual_mask_file == actual_mask_file_from_df

        # snpeff annotations should be added in headers
        reader = vcf.Reader(filename=str(actual_mutations_snpeff_file))
        assert 'ANN' in reader.infos


def test_create_fofn_file_snpeff_with_conda():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_snpeff_file = tmp_dir / 'assemblies' / 'variant-snpeff' / 'SampleA.vcf.gz'
        actual_snpeff_config = tmp_dir / 'snpeff_db' / 'snpEff.config'
        actual_mask_file = tmp_dir / 'assemblies' / 'mask' / 'SampleA.bed.gz'
        input_samples = [snpeff_input_sampleA]

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=True,
                                                      include_mlst=False, include_kmer=False)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=snpeff_reference_genome,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()
        assert actual_snpeff_config.exists()
        assert actual_mutations_snpeff_file.exists()
        assert actual_mask_file.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        assert ['Sample', 'VCF', 'Mask File'] == fofn_df.columns.tolist()

        assert 1 == len(fofn_df)
        assert ['SampleA'] == fofn_df['Sample'].tolist()
        actual_mutations_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['VCF'].tolist()[0])
        actual_mask_file_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA']['Mask File'].tolist()[0])

        assert actual_mutations_snpeff_file == actual_mutations_file_from_df
        assert actual_mask_file == actual_mask_file_from_df

        # snpeff annotations should be added in headers
        reader = vcf.Reader(filename=str(actual_mutations_snpeff_file))
        assert 'ANN' in reader.infos


def test_create_fofn_file_snpeff_reads_with_conda():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)
        actual_mutations_snpeff_file_paired = tmp_dir / 'reads' / 'paired' / 'variant-snpeff' / 'SampleA-snpeff.vcf.gz'
        actual_mutations_snpeff_file_single = tmp_dir / 'reads' / 'single' / 'variant-snpeff' / 'SampleA-single-snpeff.vcf.gz'
        actual_snpeff_config = tmp_dir / 'snpeff_db' / 'snpEff.config'
        actual_mask_file_paired = tmp_dir / 'reads' / 'paired' / 'mask' / 'SampleA-snpeff.bed.gz'
        actual_mask_file_single = tmp_dir / 'reads' / 'single' / 'mask' / 'SampleA-single-snpeff.bed.gz'
        input_samples = snpeff_reads_paired + snpeff_reads_single

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir, use_conda=True,
                                                      include_mlst=False, include_kmer=False,
                                                      reads_mincov=5)

        sample_files = pipeline_executor.create_input_sample_files(input_samples)
        results = pipeline_executor.execute(sample_files=sample_files,
                                            reference_file=snpeff_reference_genome,
                                            ncores=1)

        input_fofn = results.get_file('gdi-fofn')

        assert input_fofn.exists()
        assert actual_snpeff_config.exists()
        assert actual_mutations_snpeff_file_paired.exists()
        assert actual_mutations_snpeff_file_single.exists()
        assert actual_mask_file_paired.exists()
        assert actual_mask_file_single.exists()

        # Verify input file of file names for rest of gdi software (used as input to the indexing component)
        fofn_df = pd.read_csv(input_fofn, sep='\t')
        assert ['Sample', 'VCF', 'Mask File'] == fofn_df.columns.tolist()

        assert 2 == len(fofn_df)
        assert {'SampleA-snpeff', 'SampleA-single-snpeff'} == set(fofn_df['Sample'].tolist())
        actual_mutations_file_paired_from_df = Path(fofn_df[fofn_df['Sample'] == 'SampleA-snpeff']['VCF'].tolist()[0])
        actual_mutations_file_single_from_df = Path(
            fofn_df[fofn_df['Sample'] == 'SampleA-single-snpeff']['VCF'].tolist()[0])
        actual_mask_file_paired_from_df = Path(
            fofn_df[fofn_df['Sample'] == 'SampleA-snpeff']['Mask File'].tolist()[0])
        actual_mask_file_single_from_df = Path(
            fofn_df[fofn_df['Sample'] == 'SampleA-single-snpeff']['Mask File'].tolist()[0])

        assert actual_mutations_snpeff_file_paired == actual_mutations_file_paired_from_df
        assert actual_mutations_snpeff_file_single == actual_mutations_file_single_from_df
        assert actual_mask_file_paired == actual_mask_file_paired_from_df
        assert actual_mask_file_single == actual_mask_file_single_from_df

        # snpeff annotations should be added in headers
        reader = vcf.Reader(filename=str(actual_mutations_snpeff_file_paired))
        assert 'ANN' in reader.infos

        reader = vcf.Reader(filename=str(actual_mutations_snpeff_file_single))
        assert 'ANN' in reader.infos


def test_split_input_sequence_files_single_record():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir)
        sample_data = pipeline_executor.split_input_sequence_files([reference_file], output_dir=tmp_dir)

        expected_out1 = tmp_dir / 'reference.fasta.gz'

        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == sample_data.columns.tolist()
        assert 1 == len(sample_data)
        assert ['reference'] == sample_data['Sample'].tolist()
        assert [expected_out1] == sample_data['Assemblies'].tolist()
        assert sample_data['Reads1'].isna().all()
        assert sample_data['Reads2'].isna().all()

        assert expected_out1.exists()

        sf = SequenceFile(expected_out1)
        name, records = sf.parse_sequence_file()
        assert 5180 == len(records[0])


def test_split_input_sequence_files_multiple_records():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir)
        sample_data = pipeline_executor.split_input_sequence_files([reference_file,
                                                                    reference_file_5000_snpeff_2], output_dir=tmp_dir)
        sample_data = sample_data.sort_values('Sample')

        expected_out1 = tmp_dir / 'CP001602.2.gbk.gz'
        expected_out2 = tmp_dir / 'NC_011083.1.gbk.gz'
        expected_out3 = tmp_dir / 'reference.fasta.gz'

        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == sample_data.columns.tolist()
        assert 3 == len(sample_data)
        assert ['CP001602.2', 'NC_011083.1', 'reference'] == sample_data['Sample'].tolist()
        assert [expected_out1, expected_out2, expected_out3] == sample_data['Assemblies'].tolist()
        assert sample_data['Reads1'].isna().all()
        assert sample_data['Reads2'].isna().all()

        assert expected_out1.exists()
        assert expected_out2.exists()
        assert expected_out3.exists()

        sf1 = SequenceFile(expected_out1)
        name, records = sf1.parse_sequence_file()
        assert 5000 == len(records[0])

        sf2 = SequenceFile(expected_out1)
        name, records = sf2.parse_sequence_file()
        assert 5000 == len(records[0])

        sf3 = SequenceFile(expected_out3)
        name, records = sf3.parse_sequence_file()
        assert 5180 == len(records[0])


def test_split_input_sequence_files_multiple_records_subsample():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        samples = {'CP001602.2', 'reference'}

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir)
        sample_data = pipeline_executor.split_input_sequence_files([reference_file,
                                                                    reference_file_5000_snpeff_2],
                                                                   output_dir=tmp_dir,
                                                                   samples=samples)
        sample_data = sample_data.sort_values('Sample')

        expected_out1 = tmp_dir / 'CP001602.2.gbk.gz'
        unexpected_out2 = tmp_dir / 'NC_011083.1.gbk.gz'
        expected_out3 = tmp_dir / 'reference.fasta.gz'

        assert ['Sample', 'Assemblies', 'Reads1', 'Reads2'] == sample_data.columns.tolist()
        assert 2 == len(sample_data)
        assert ['CP001602.2', 'reference'] == sample_data['Sample'].tolist()
        assert [expected_out1, expected_out3] == sample_data['Assemblies'].tolist()
        assert sample_data['Reads1'].isna().all()
        assert sample_data['Reads2'].isna().all()

        assert expected_out1.exists()
        assert not unexpected_out2.exists()
        assert expected_out3.exists()

        sf1 = SequenceFile(expected_out1)
        name, records = sf1.parse_sequence_file()
        assert 5000 == len(records[0])

        sf3 = SequenceFile(expected_out3)
        name, records = sf3.parse_sequence_file()
        assert 5180 == len(records[0])


def test_select_random_samples():
    with TemporaryDirectory() as tmp_dir_str:
        tmp_dir = Path(tmp_dir_str)

        samples = {'CP001602.2', 'reference', 'NC_011083.1'}

        pipeline_executor = SnakemakePipelineExecutor(working_directory=tmp_dir)

        # Test select 1 sample
        subsamples = pipeline_executor.select_random_samples([reference_file, reference_file_5000_snpeff_2],
                                                             number_samples=1)
        assert 1 == len(subsamples)
        assert subsamples.issubset(samples)

        # Test select 2 samples
        subsamples = pipeline_executor.select_random_samples([reference_file, reference_file_5000_snpeff_2],
                                                             number_samples=2)
        assert 2 == len(subsamples)
        assert subsamples.issubset(samples)

        # Test select 3 samples
        subsamples = pipeline_executor.select_random_samples([reference_file, reference_file_5000_snpeff_2],
                                                             number_samples=3)
        assert 3 == len(subsamples)
        assert subsamples.issubset(samples)

        # Test select 33% of samples
        subsamples = pipeline_executor.select_random_samples([reference_file, reference_file_5000_snpeff_2],
                                                             number_samples=0.33)
        assert 1 == len(subsamples)
        assert subsamples.issubset(samples)

        # Test select 66% of samples
        subsamples = pipeline_executor.select_random_samples([reference_file, reference_file_5000_snpeff_2],
                                                             number_samples=0.66)
        assert 2 == len(subsamples)
        assert subsamples.issubset(samples)



import pytest

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.db import NucleotideVariantsSamples, MLSTAllelesSamples


def test_update_sample_ids_both_overlap():
    s1 = SampleSet([1, 2])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    s2 = SampleSet([2, 3])
    v2 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s2)

    v1.update_sample_ids(v2)

    # v1 should only have sample_ids updated
    assert v1.id == 'ecoli:abc:1'
    assert v1.sla == 'ecoli:abc:1'
    assert set(v1.sample_ids) == {1, 2, 3}

    # v2 should not be changed
    assert v2.id == 'ecoli:abc:1'
    assert set(v2.sample_ids) == {2, 3}


def test_update_sample_ids_disjoint():
    s1 = SampleSet([1, 2])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    s2 = SampleSet([4, 10])
    v2 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s2)

    v1.update_sample_ids(v2)

    # v1 should only have sample_ids updated
    assert v1.id == 'ecoli:abc:1'
    assert v1.sla == 'ecoli:abc:1'
    assert set(v1.sample_ids) == {1, 2, 4, 10}

    # v2 should not be changed
    assert v2.id == 'ecoli:abc:1'
    assert set(v2.sample_ids) == {4, 10}


def test_update_sample_ids_feature_mismatch():
    s1 = SampleSet([1, 2])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    s2 = SampleSet([2, 3])
    v2 = NucleotideVariantsSamples(spdi='ref:10:1:A', var_type='SNP', sample_ids=s2)

    with pytest.raises(Exception) as execinfo:
        v1.update_sample_ids(v2)

    assert 'Cannot merge other' in str(execinfo.value)
    assert 'since it is not of type' in str(execinfo.value)


def test_update_sample_ids_feature_id_mismatch():
    s1 = SampleSet([1, 2])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    s2 = SampleSet([2, 3])
    v2 = MLSTAllelesSamples(sla='ecoli-n:abc:1', sample_ids=s2)

    with pytest.raises(Exception) as execinfo:
        v1.update_sample_ids(v2)

    assert 'Cannot merge other' in str(execinfo.value)
    assert 'since identifiers are not equal' in str(execinfo.value)


def test_query_id():
    s1 = SampleSet([1, 2])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    assert v1.query_id == 'mlst:ecoli:abc:1'


def test_samples_is_unknown():
    s1 = SampleSet([1])
    v1 = MLSTAllelesSamples(sla='ecoli:abc:1', sample_ids=s1)

    s2 = SampleSet([2])
    v2 = MLSTAllelesSamples(sla='ecoli:abc:?', sample_ids=s2)

    # v1 should not be unknown
    assert v1.id == 'ecoli:abc:1'
    assert not v1.is_unknown

    # v2 should be unknown
    assert v2.id == 'ecoli:abc:?'
    assert v2.is_unknown



from __future__ import annotations

from typing import Iterable, Generator, Union, Set, List

from pyroaring import BitMap


class SampleSet:

    def __init__(self, sample_ids: Iterable[int] = None, existing_bitmap: BitMap = None):
        if sample_ids is None and existing_bitmap is None:
            raise Exception('Both sample_ids and existing_bitmap are unset. One of them must be set.')
        elif sample_ids is not None and existing_bitmap is not None:
            raise Exception('Both sample_ids and existing_bitmap are set. Only one of them can be set')
        elif sample_ids is not None:
            self._bitmap = BitMap(sample_ids)
        else:
            self._bitmap = existing_bitmap

    def intersection(self, other: Union[Set[int], SampleSet]) -> SampleSet:
        if other is None:
            raise Exception('Cannot intersect other[None]')
        elif isinstance(other, AllSampleSet):
            return self
        elif isinstance(other, SampleSet):
            return SampleSet(existing_bitmap=self._bitmap.intersection(other._bitmap))
        elif isinstance(other, set):
            return SampleSet(self._bitmap.intersection(BitMap(other)))
        else:
            raise Exception(f'Cannot intersect other of type [{type(other)}]')

    def union(self, other: Union[Set[int], SampleSet]) -> SampleSet:
        if other is None:
            raise Exception('Cannot union other=[None]')
        elif isinstance(other, AllSampleSet):
            return other
        elif isinstance(other, SampleSet):
            return SampleSet(existing_bitmap=self._bitmap.union(other._bitmap))
        elif isinstance(other, set):
            return SampleSet(self._bitmap.union(BitMap(other)))
        else:
            raise Exception(f'Cannot union other of type [{type(other)}]')

    def minus(self, other: Union[Set[int], SampleSet, List[SampleSet]]) -> SampleSet:
        if other is None:
            raise Exception('Cannot union other=[None]')
        elif isinstance(other, AllSampleSet):
            return SampleSet.create_empty()
        elif isinstance(other, SampleSet):
            return SampleSet(existing_bitmap=(self._bitmap - other._bitmap))
        elif isinstance(other, set):
            return SampleSet(existing_bitmap=(self._bitmap - BitMap(other)))
        elif isinstance(other, list):
            difference_bitmap = self._bitmap
            for s in other:
                if isinstance(s, SampleSet):
                    difference_bitmap = difference_bitmap - s._bitmap
                else:
                    raise Exception(f'Got invalid type=[{type(s)}] in input other=[{other}]. '
                                    f'Expected a list of SampleSet')
            return SampleSet(existing_bitmap=difference_bitmap)
        else:
            raise Exception(f'Cannot union other of type [{type(other)}]')

    def jaccard_index(self, other: Union[Set[int], SampleSet]) -> float:
        if other is None:
            raise Exception('Cannot calculate jaccard with other=[None]')
        elif isinstance(other, AllSampleSet):
            raise NotImplementedError('Jaccard with AllSampleSet is not implemented')
        elif isinstance(other, SampleSet):
            return self._bitmap.jaccard_index(other._bitmap)
        elif isinstance(other, set):
            return self._bitmap.jaccard_index(SampleSet(other)._bitmap)
        else:
            raise Exception(f'Cannot union other of type [{type(other)}]')

    def is_empty(self) -> bool:
        return len(self._bitmap) == 0

    @staticmethod
    def from_bytes(data: bytes) -> SampleSet:
        bitmap = BitMap.deserialize(data)
        return SampleSet(existing_bitmap=bitmap)

    @classmethod
    def create_empty(cls):
        return SampleSet(existing_bitmap=BitMap())

    @classmethod
    def create_all(cls):
        return AllSampleSet()

    def get_bytes(self) -> bytes:
        return self._bitmap.serialize()

    def __iter__(self) -> Generator[int, None, None]:
        yield from self._bitmap

    def __contains__(self, value: int) -> bool:
        return value in self._bitmap

    def __len__(self) -> int:
        return len(self._bitmap)

    def __repr__(self):
        return (f'<SampleSet(size={len(self._bitmap)})>')


class AllSampleSet(SampleSet):
    # Roaring bitmaps store 32-bit integers, so max size is all 32-bit ints
    MAX_VALUE = 2 ** 32

    def __init__(self):
        super().__init__(sample_ids=[])

    def intersection(self, other: Union[Set[int], SampleSet]) -> SampleSet:
        return other

    def union(self, other: Union[Set[int], SampleSet]) -> SampleSet:
        return self

    def is_empty(self) -> bool:
        return False

    def minus(self, other: Union[Set[int], SampleSet, List[SampleSet]]) -> SampleSet:
        raise Exception('Cannot subtract anything from AllSampleSet')

    def get_bytes(self) -> bytes:
        raise Exception('Cannot serialize AllSampleSet')

    def __iter__(self) -> Generator[int, None, None]:
        yield from range(self.MAX_VALUE)

    def __contains__(self, value: int) -> bool:
        return value < self.MAX_VALUE

    def __len__(self) -> int:
        return self.MAX_VALUE

    def __repr__(self):
        return (f'<AllSampleSet(size={len(self._bitmap)})>')



import tempfile
from pathlib import Path
from typing import Optional

from pybedtools import BedTool

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.NucleotideSampleDataBedMask import NucleotideSampleDataBedMask
from genomics_data_index.test.integration import variation_dir

sampleA_mask_file = variation_dir / 'SampleA.bed.gz'
sampleA_vcf = variation_dir / 'SampleA.vcf.gz'


def create_sample_data(name: str, vcf: Path, mask: Optional[Path]) -> NucleotideSampleData:
    return NucleotideSampleDataBedMask.create(sample_name=name, vcf_file=vcf,
                                              mask_bed_file=mask)


def test_preprocess_mask():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        sample_data = NucleotideSampleDataBedMask.create(sample_name='SampleA', vcf_file=sampleA_vcf,
                                                         mask_bed_file=sampleA_mask_file)
        assert not sample_data.is_preprocessed()

        data_preprocessed = sample_data.preprocess(tmp_dir)

        assert data_preprocessed.is_preprocessed()
        vcf_file, vcf_file_index = data_preprocessed.get_vcf_file()
        assert tmp_dir == data_preprocessed.get_mask_file().parent
        assert tmp_dir == vcf_file.parent
        assert tmp_dir == vcf_file_index.parent
        assert 70 == len(data_preprocessed.get_mask())


def test_preprocess_mask_empty():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        sample_data = NucleotideSampleDataBedMask.create(sample_name='SampleA', vcf_file=sampleA_vcf,
                                                         mask_bed_file=None)
        assert not sample_data.is_preprocessed()

        data_preprocessed = sample_data.preprocess(tmp_dir)

        assert data_preprocessed.is_preprocessed()
        vcf_file, vcf_file_index = data_preprocessed.get_vcf_file()
        assert tmp_dir == data_preprocessed.get_mask_file().parent
        assert tmp_dir == vcf_file.parent
        assert tmp_dir == vcf_file_index.parent

        assert data_preprocessed.get_mask().is_empty()


def test_preprocess_other_mask():
    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_dir = Path(tmp_dir)
        initial_mask = tmp_dir / 'initial.bed.gz'
        BedTool([('NC_011083', 0, 100)]).saveas(initial_mask)
        sample_data = NucleotideSampleDataBedMask.create(sample_name='SampleA', vcf_file=sampleA_vcf,
                                                         mask_bed_file=initial_mask)
        assert not sample_data.is_preprocessed()

        data_preprocessed = sample_data.preprocess(tmp_dir)

        assert data_preprocessed.is_preprocessed()
        vcf_file, vcf_file_index = data_preprocessed.get_vcf_file()
        assert tmp_dir == data_preprocessed.get_mask_file().parent
        assert tmp_dir == vcf_file.parent
        assert tmp_dir == vcf_file_index.parent

        assert 100 == len(data_preprocessed.get_mask())



import logging
from pathlib import Path
from typing import List, Set, Any, Dict, cast, Union

import pandas as pd
import sqlalchemy

from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.io.FeaturesReader import FeaturesReader
from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage
from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mutation.VariationFile import VariationFile
from genomics_data_index.storage.io.mutation.VcfVariantsReader import VcfVariantsReader
from genomics_data_index.storage.model import NUCLEOTIDE_UNKNOWN_TYPE
from genomics_data_index.storage.model.QueryFeatureHGVSGN import QueryFeatureHGVSGN
from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI
from genomics_data_index.storage.model.db import NucleotideVariantsSamples, SampleNucleotideVariation, Sample, \
    FeatureSamples
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service import SQLQueryInBatcherDict
from genomics_data_index.storage.service.FeatureService import FeatureService
from genomics_data_index.storage.service.ReferenceService import ReferenceService
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.util.SamplesProgressLogger import SamplesProgressLogger

logger = logging.getLogger(__name__)


class VariationService(FeatureService):
    MUTATION_TYPES = ['snp', 'indel', 'all', 'other']
    MUTATION_ID_TYPES = ['spdi_ref', 'spdi']

    def __init__(self, database_connection: DatabaseConnection, variation_dir: Path,
                 reference_service: ReferenceService, sample_service: SampleService,
                 index_unknown_missing: bool, sql_select_limit: int):
        super().__init__(database_connection=database_connection,
                         features_dir=variation_dir,
                         sample_service=sample_service)
        self._reference_service = reference_service
        self._index_unknown_missing = index_unknown_missing
        self._sql_select_limit = sql_select_limit

    def _reference_sequence_names(self, reference_name: str) -> List[str]:
        return list(self._reference_service.get_reference_sequences(reference_name).keys())

    def _query_include_unknown(self, query, include_present: bool = True, include_unknown: bool = False):
        if not include_present:
            if include_unknown:
                return query.filter(NucleotideVariantsSamples.var_type == NUCLEOTIDE_UNKNOWN_TYPE)
            else:
                # Should return a query that gives no results
                return query.filter(sqlalchemy.sql.false())
        elif not include_unknown:
            return query.filter(NucleotideVariantsSamples.var_type != NUCLEOTIDE_UNKNOWN_TYPE)
        else:
            return query

    def count_on_reference(self, reference_name: str, include_unknown: bool = False) -> int:
        reference_sequence_names = self._reference_sequence_names(reference_name)

        query = self._connection.get_session().query(NucleotideVariantsSamples) \
            .filter(NucleotideVariantsSamples.sequence.in_(reference_sequence_names))

        return self._query_include_unknown(query, include_unknown=include_unknown).count()

    def mutation_counts_on_reference(self, reference_name: str, include_unknown: bool = False) -> Dict[str, int]:
        reference_sequence_names = self._reference_sequence_names(reference_name)

        query = self._connection.get_session().query(NucleotideVariantsSamples) \
            .filter(NucleotideVariantsSamples.sequence.in_(reference_sequence_names))
        mutations = self._query_include_unknown(query, include_unknown=include_unknown).all()

        return {m.spdi: len(m.sample_ids) for m in mutations}

    def get_features(self, include_present: bool = True,
                     include_unknown: bool = False,
                     id_type: str = 'spdi') -> Dict[str, NucleotideVariantsSamples]:
        query = self._connection.get_session().query(NucleotideVariantsSamples)

        if not include_present:
            if include_unknown:
                query = query.filter(NucleotideVariantsSamples.var_type == NUCLEOTIDE_UNKNOWN_TYPE)
            else:
                # Here, I'm not including unknown or present, so don't query database
                return dict()
        elif not include_unknown:
            query = query.filter(NucleotideVariantsSamples.var_type != NUCLEOTIDE_UNKNOWN_TYPE)

        features = query.all()
        if id_type == 'spdi_ref':
            feature_ids_db = {f.id: f for f in features}
            feature_ids_translated = self._reference_service.translate_spdi(feature_ids_db.keys(), to=id_type)
            return {feature_ids_translated[fid]: feature_ids_db[fid] for fid in feature_ids_db}
        elif id_type == 'spdi':
            return {f.id: f for f in features}
        else:
            raise Exception(f'Unknown value for id_type={id_type}. Must be one of {self.MUTATION_ID_TYPES}')

    def get_variants_on_reference(self, reference_name: str, include_present: bool = True,
                                  include_unknown: bool = False) -> Dict[str, NucleotideVariantsSamples]:
        reference_sequence_names = self._reference_sequence_names(reference_name)

        query = self._connection.get_session().query(NucleotideVariantsSamples) \
            .filter(NucleotideVariantsSamples.sequence.in_(reference_sequence_names))

        mutations = self._query_include_unknown(query, include_present=include_present,
                                                include_unknown=include_unknown).all()

        return {m.spdi: m for m in mutations}

    def count_mutations_in_sample_ids_dataframe(self, sample_ids: Union[SampleSet, List[int]],
                                                ncores: int = 1,
                                                batch_size: int = 50,
                                                mutation_type: str = 'all',
                                                include_unknown: bool = False) -> pd.DataFrame:
        if include_unknown:
            raise Exception(f'support for include_unknown is not implemented')

        if mutation_type == 'all':
            include_expression = None
        elif mutation_type == 'snp':
            include_expression = 'TYPE="SNP"'
        elif mutation_type == 'indel':
            include_expression = 'TYPE="INDEL"'
        elif mutation_type == 'other':
            include_expression = 'TYPE="OTHER"'
        else:
            raise Exception(f'Unsupported option mutation_type=[{mutation_type}]. Must be one of {self.MUTATION_TYPES}')

        if isinstance(sample_ids, SampleSet):
            sample_ids = list(sample_ids)

        sample_nucleotide_variation = self._connection.get_session().query(SampleNucleotideVariation) \
            .filter(SampleNucleotideVariation.sample_id.in_(sample_ids)) \
            .all()

        nucleotide_variants_files = [snv.nucleotide_variants_file for snv in sample_nucleotide_variation]
        mutation_df = VariationFile.union_all_files(nucleotide_variants_files,
                                                    include_expression=include_expression,
                                                    batch_size=batch_size,
                                                    ncores=ncores)

        mutation_df = mutation_df.rename(columns={
            'ID': 'Mutation',
            'CHROM': 'Sequence',
            'POS': 'Position',
            'REF': 'Deletion',
            'ALT': 'Insertion',
            'COUNT': 'Count',
        })

        return mutation_df.set_index('Mutation')

    def append_mutation_annotations(self, features_df: pd.DataFrame) -> pd.DataFrame:
        """
        Adds annotations to the mutations stored within the passed dataframe.
        :param features_df: The dataframe to add annotations.
                            Assumes the index of the dataframe contains the SPDI identifier.
        :return: A new dataframe with mutation annotations.
        """
        if features_df.index.name != 'Mutation':
            raise Exception(f'Index name not equal to "Mutation" for features_df={features_df}')

        mutation_ids = features_df.index.tolist()
        query_features = [QueryFeatureMutationSPDI(i) for i in mutation_ids]
        id_to_nucleotide_variants_samples: Dict[str, NucleotideVariantsSamples] = \
            self._sample_service.get_variants_samples_by_variation_features(query_features)

        annotation_data = []
        for mutation_id in id_to_nucleotide_variants_samples:
            variants_samples = id_to_nucleotide_variants_samples[mutation_id]

            if variants_samples.id_hgvs_c is not None:
                id_hgvs_gn_c = QueryFeatureHGVSGN.create(sequence_name=variants_samples.sequence,
                                                         gene=variants_samples.annotation_gene_name,
                                                         variant=variants_samples.annotation_hgvs_c).id
            else:
                id_hgvs_gn_c = pd.NA

            if variants_samples.id_hgvs_p is not None:
                id_hgvs_gn_p = QueryFeatureHGVSGN.create(sequence_name=variants_samples.sequence,
                                                         gene=variants_samples.annotation_gene_name,
                                                         variant=variants_samples.annotation_hgvs_p).id
            else:
                id_hgvs_gn_p = pd.NA

            annotation_data.append([mutation_id,
                                    variants_samples.annotation if variants_samples.annotation is not None else pd.NA,
                                    variants_samples.annotation_impact if variants_samples.annotation_impact is not None else pd.NA,
                                    variants_samples.annotation_gene_name if variants_samples.annotation_gene_name is not None else pd.NA,
                                    variants_samples.annotation_gene_id if variants_samples.annotation_gene_id is not None else pd.NA,
                                    variants_samples.annotation_feature_type if variants_samples.annotation_feature_type is not None else pd.NA,
                                    variants_samples.annotation_transcript_biotype if variants_samples.annotation_transcript_biotype is not None else pd.NA,
                                    variants_samples.annotation_hgvs_c if variants_samples.annotation_hgvs_c is not None else pd.NA,
                                    variants_samples.annotation_hgvs_p if variants_samples.annotation_hgvs_p is not None else pd.NA,
                                    variants_samples.id_hgvs_c if variants_samples.id_hgvs_c is not None else pd.NA,
                                    variants_samples.id_hgvs_p if variants_samples.id_hgvs_p is not None else pd.NA,
                                    id_hgvs_gn_c,
                                    id_hgvs_gn_p,
                                    ])

        annotation_df = pd.DataFrame(data=annotation_data,
                                     columns=['Mutation',
                                              'Annotation',
                                              'Annotation_Impact',
                                              'Gene_Name',
                                              'Gene_ID',
                                              'Feature_Type',
                                              'Transcript_BioType',
                                              'HGVS.c',
                                              'HGVS.p',
                                              'ID_HGVS.c',
                                              'ID_HGVS.p',
                                              'ID_HGVS_GN.c',
                                              'ID_HGVS_GN.p',
                                              ]).set_index('Mutation')

        return features_df.merge(annotation_df, how='left', left_index=True, right_index=True)

    def get_variants_ordered(self, sequence_name: str, type: str = 'SNP') -> List[NucleotideVariantsSamples]:
        return self._connection.get_session().query(NucleotideVariantsSamples) \
            .filter(NucleotideVariantsSamples.sequence == sequence_name) \
            .filter(NucleotideVariantsSamples.var_type == type) \
            .order_by(NucleotideVariantsSamples.position) \
            .all()

    def get_sample_nucleotide_variation(self, sample_names: List[str]) -> List[SampleNucleotideVariation]:
        return self._connection.get_session().query(SampleNucleotideVariation) \
            .join(SampleNucleotideVariation.sample) \
            .filter(Sample.name.in_(sample_names)) \
            .all()

    def _create_feature_identifier(self, features_df: pd.DataFrame) -> str:
        return NucleotideVariantsSamples.to_spdi(
            sequence_name=features_df['CHROM'],
            position=features_df['POS'],
            ref=features_df['REF'],
            alt=features_df['ALT']
        )

    def _get_sample_id_series(self, features_df: pd.DataFrame, sample_name_ids: Dict[str, int]) -> pd.Series:
        return features_df.apply(lambda x: sample_name_ids[x['SAMPLE']], axis='columns')

    def _create_feature_object(self, features_df: pd.DataFrame) -> FeatureSamples:
        return NucleotideVariantsSamples(spdi=features_df['_FEATURE_ID'],
                                         var_type=features_df['TYPE'],
                                         sample_ids=features_df['_SAMPLE_ID'],
                                         annotation=features_df['ANN.Annotation'],
                                         annotation_impact=features_df['ANN.Annotation_Impact'],
                                         annotation_gene_name=features_df['ANN.Gene_Name'],
                                         annotation_gene_id=features_df['ANN.Gene_ID'],
                                         annotation_feature_type=features_df['ANN.Feature_Type'],
                                         annotation_transcript_biotype=features_df['ANN.Transcript_BioType'],
                                         annotation_hgvs_c=features_df['ANN.HGVS.c'],
                                         annotation_hgvs_p=features_df['ANN.HGVS.p'])

    def get_correct_data_package(self) -> Any:
        return NucleotideSampleDataPackage

    def get_correct_sample_data(self) -> Any:
        return NucleotideSampleData

    def aggregate_feature_column(self) -> Dict[str, Any]:
        return {'TYPE': 'first', '_SAMPLE_ID': SampleSet,
                'ANN.Annotation': 'first', 'ANN.Annotation_Impact': 'first',
                'ANN.Gene_Name': 'first', 'ANN.Gene_ID': 'first', 'ANN.Feature_Type': 'first',
                'ANN.Transcript_BioType': 'first', 'ANN.HGVS.c': 'first', 'ANN.HGVS.p': 'first'}

    def _modify_df_types(self, features_df: pd.DataFrame) -> pd.DataFrame:
        features_df = features_df.copy()
        columns = ['ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
                   'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p']

        # Convert NA to None to properly save in database
        features_df[columns] = features_df[columns].apply(lambda x: x.astype('object'))
        features_df[columns] = features_df[columns].where(pd.notnull(features_df[columns]), None)

        return features_df

    def check_samples_have_features(self, sample_names: Set[str], feature_scope_name: str) -> bool:
        samples_with_variants = {sample.name for sample in
                                 self._sample_service.get_samples_with_variants(feature_scope_name)}
        return len(samples_with_variants.intersection(sample_names)) != 0

    def _verify_correct_feature_scope(self, feature_scope_name: str) -> None:
        if feature_scope_name is None:
            raise Exception('feature_scope_name must not be None')

    def build_sample_feature_object(self, sample: Sample, sample_data: SampleData, feature_scope_name: str) -> Any:
        nucleotide_sample_files = cast(NucleotideSampleData, sample_data)

        reference = self._reference_service.find_reference_genome(feature_scope_name)

        sample_nucleotide_variation = SampleNucleotideVariation(reference=reference)
        vcf_file, vcf_index = nucleotide_sample_files.get_vcf_file()
        sample_nucleotide_variation.nucleotide_variants_file = vcf_file
        sample_nucleotide_variation.sample = sample
        sample_nucleotide_variation.masked_regions_file = nucleotide_sample_files.get_mask_file()

        return sample_nucleotide_variation

    def _create_persisted_features_reader(self, sample_data_dict: Dict[str, SampleData],
                                          data_package: SampleDataPackage) -> FeaturesReader:
        sample_data_dict = cast(Dict[str, NucleotideSampleData], sample_data_dict)
        data_package = cast(NucleotideSampleDataPackage, data_package)
        index_unknown_missing = self._index_unknown_missing and data_package.index_unknown_missing()
        if not index_unknown_missing:
            logger.debug(f'index_unknown_missing={index_unknown_missing} so will not '
                         'index missing/unknown positions')
        else:
            logger.debug(f'index_unknown_missing={index_unknown_missing}')
        progress_logger = SamplesProgressLogger(stage_name='Index', stage_number=2, total_samples=len(sample_data_dict))
        return VcfVariantsReader(sample_data_dict, include_masked_regions=index_unknown_missing,
                                 progress_logger=progress_logger,
                                 variants_processor_factory=data_package.get_variants_processor_factory())

    def read_index(self, feature_ids: Union[List[str], Set[str]]) -> Dict[str, FeatureSamples]:
        if isinstance(feature_ids, set):
            feature_ids = list(feature_ids)

        query_batcher = SQLQueryInBatcherDict(in_data=feature_ids, batch_size=self._sql_select_limit)

        def handle_batch(feature_ids_batch: List[str]) -> Dict[str, FeatureSamples]:
            feature_samples = self._connection.get_session().query(NucleotideVariantsSamples) \
                .filter(NucleotideVariantsSamples._spdi.in_(feature_ids_batch)) \
                .all()
            return {f.id: f for f in feature_samples}

        return query_batcher.process(handle_batch)



from pathlib import Path

import pandas as pd
import vcf


def read_vcf_df(file: Path) -> pd.DataFrame:
    reader = vcf.Reader(filename=str(file))
    df = pd.DataFrame([vars(r) for r in reader])
    df['TYPE'] = df['INFO'].apply(lambda x: x['TYPE'][0] if 'TYPE' in x else pd.NA)

    return df



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import List, Union

import pandas as pd
from ete3 import Tree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.features.MLSTFeaturesComparator import MLSTFeaturesComparator
from genomics_data_index.api.query.features.MutationFeaturesFromIndexComparator import \
    MutationFeaturesFromIndexComparator
from genomics_data_index.api.query.impl.DataFrameSamplesQuery import DataFrameSamplesQuery
from genomics_data_index.api.query.impl.SamplesQueryIndex import SamplesQueryIndex
from genomics_data_index.api.query.impl.TreeSamplesQueryFactory import TreeSamplesQueryFactory
from genomics_data_index.configuration.Project import Project
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.service import EntityExistsError

logger = logging.getLogger(__name__)


class GenomicsDataIndex:
    """
    The root class for the Python API. Used to connect to a genomics index and build query objects.
    """

    QUERY_UNIVERSE = ['all', 'mutations', 'mutations_experimental', 'dataframe']
    MUTATION_ID_TYPES = ['spdi_ref', 'spdi']
    FEAUTRE_KINDS = ['mutations', 'mlst']

    def __init__(self, connection: DataIndexConnection):
        """
        Builds a new GenomicsDataIndex.
        :param connection: The database connection to use.

        :return: A new GenomicsDataIndex.
        """
        self._connection = connection
        self._samples_count = self.count_samples()

    @property
    def connection(self) -> DataIndexConnection:
        """
        Gets the database connection object.

        :return: The database connection object.
        """
        return self._connection

    def count_samples(self) -> int:
        """
        Counts the samples stored in this genomics index.

        :return: The count of samples stored in this genomics index.
        """
        return self._connection.sample_service.count_samples()

    def sample_names(self) -> List[str]:
        """
        Gets a list of all sample names stored in this index.

        :return: A list of all sample names stored in this index.
        """
        return [s.name for s in self._connection.sample_service.get_samples()]

    def count_references(self) -> int:
        """
        Counts the number of reference genomes stored in this genomics index.

        :return: The count of reference genomes stored in this genomics index.
        """
        return self._connection.reference_service.count_reference_genomes()

    def reference_names(self) -> List[str]:
        """
        Gets a list of the names of the reference genomes used in this index.
        :return: A list of all reference genome names.
        """
        return [r.name for r in self._connection.reference_service.get_reference_genomes()]

    def mlst_schemes(self) -> List[str]:
        """
        Gets a list of the MLST schemes used in this index.
        :return: A list of all MLST scheme names.
        """
        return [r.name for r in self._connection.mlst_service.get_mlst_schemes()]

    def reference_tree(self, reference_name: str) -> Tree:
        """
        Gets a tree associated with a reference genome.
        :param reference_name: The name of the reference genome to get a tree for.
        :return: The tree (if any) associated with the reference genome.
        """
        reference_genome = self._connection.reference_service.find_reference_genome(reference_name)
        if reference_genome.has_tree():
            return reference_genome.tree
        else:
            raise EntityExistsError(f'reference_genome=[{reference_genome}] does not have a tree')

    def count_mutations(self, reference_genome: str, include_unknown: bool = False) -> int:
        """
        Counts all mutations indexed relative to the given reference genome.

        :param reference_genome: The reference genome.
        :param include_unknown: Whether or not unknown mutations should be included.

        :return: A count of all mutations relative to the given reference genome.
        """
        return self._connection.variation_service.count_on_reference(reference_genome,
                                                                     include_unknown=include_unknown)

    def mutations_summary(self, reference_name: str, id_type: str = 'spdi_ref', include_present_features: bool = True,
                          include_unknown_features: bool = False,
                          include_unknown_samples: bool = False,
                          include_unknown_no_present_samples: bool = False,
                          ignore_annotations: bool = False) -> pd.DataFrame:
        """
        Summarizes all mutations stored in this index relative to a string for the passed scope.
        Shorthand for features_summary(kind='mutations', ...)

        :param reference_name: The reference genome name.
        :param id_type: The type of identifier to use.
        :param include_present_features: Whether or not mutation features present in this index (i.e., not unknown/missing)
                                should be included.
        :param include_unknown_features: Whether or not unknown mutations should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :param ignore_annotations: Whether or not mutation annotations should be ignored.
        :return: A summary of all mutations in this index as a DataFrame.
        """
        return self.features_summary(kind='mutations', scope=reference_name,
                                     include_present_features=include_present_features,
                                     include_unknown_features=include_unknown_features,
                                     include_unknown_samples=include_unknown_samples,
                                     include_unknown_no_present_samples=include_unknown_no_present_samples,
                                     id_type=id_type, ignore_annotations=ignore_annotations)

    def mlst_summary(self, scheme_name: str, locus: str = None, include_present_features: bool = True,
                     include_unknown_features: bool = False,
                     include_unknown_samples: bool = False,
                     include_unknown_no_present_samples: bool = False) -> pd.DataFrame:
        """
        Summarizes all MLST alleles stored in this index relative to the passed scheme name.
        Shorthand for features_summary(kind='mlst', ...)

        :param scheme_name: The MLST scheme to summarize.
        :param locus: The locus id to restrict summaries to. Defaults to all locus IDs.
        :param include_present_features: Whether or not MLST features present in this index (i.e., not unknown/missing)
                                should be included.
        :param include_unknown_features: Whether or not unknown MLST alleles should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :return: A summary of all MLST alleles in this index as a DataFrame.
        """
        return self.features_summary(kind='mlst', scope=scheme_name, include_present_features=include_present_features,
                                     include_unknown_features=include_unknown_features,
                                     include_unknown_samples=include_unknown_samples,
                                     include_unknown_no_present_samples=include_unknown_no_present_samples,
                                     locus=locus)

    def features_summary(self, kind: str = 'mutations', scope: str = None,
                         include_present_features: bool = True, include_unknown_features: bool = False,
                         include_unknown_samples: bool = False, include_unknown_no_present_samples: bool = False,
                         **kwargs) -> pd.DataFrame:
        """
        Summarizes all features stored in this index relative to a string for the passed scope.

        :param kind: The kind of feature (e.g., 'mutations' or 'mlst').
        :param scope: The scope (e.g., reference genome or MLST scheme).
        :param include_present_features: Whether or not features present in this index (i.e., not unknown/missing)
                                should be included.
        :param include_unknown_features: Whether or not unknown/missing features should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :return: A summary of all features in this index as a DataFrame.
        """
        if kind == 'mutations' or kind == 'mutation':
            return self._mutations_summary_internal(reference_name=scope,
                                                    include_present_features=include_present_features,
                                                    include_unknown_features=include_unknown_features,
                                                    include_unknown_samples=include_unknown_samples,
                                                    include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                    **kwargs)
        elif kind == 'mlst':
            return self._mlst_summary_internal(scheme_name=scope, include_present_features=include_present_features,
                                               include_unknown_features=include_unknown_features,
                                               include_unknown_samples=include_unknown_samples,
                                               include_unknown_no_present_samples=include_unknown_no_present_samples,
                                               **kwargs)
        else:
            raise Exception(f'Unknown value for kind=[{kind}]. Must be one of {self.FEAUTRE_KINDS}.')

    def _mlst_summary_internal(self, scheme_name: str, locus: str = None, include_present_features: bool = True,
                               include_unknown_features: bool = False,
                               include_unknown_samples: bool = False,
                               include_unknown_no_present_samples: bool = False
                               ) -> pd.DataFrame:
        """
        Summarizes all MLST alleles stored in this index relative to the passed scheme name.

        :param scheme_name: The MLST scheme to summarize.
        :param locus: The locus id to restrict summaries to. Defaults to all locus IDs.
        :param include_present_features: Whether or not MLST features present in this index (i.e., not unknown/missing)
                                should be included.
        :param include_unknown_features: Whether or not unknown MLST alleles should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :return: A summary of all MLST alleles in this index as a DataFrame.
        """
        features_summarizier = MLSTFeaturesComparator(connection=self._connection,
                                                      scheme=scheme_name,
                                                      locus=locus,
                                                      include_unknown=include_unknown_features,
                                                      include_present=include_present_features,
                                                      include_unknown_samples=include_unknown_samples,
                                                      include_unknown_no_present_samples=include_unknown_no_present_samples)
        return features_summarizier.summary(SampleSet.create_all())

    def _mutations_summary_internal(self, reference_name: str, id_type: str = 'spdi_ref',
                                    include_present_features: bool = True,
                                    include_unknown_features: bool = False,
                                    include_unknown_samples: bool = False,
                                    include_unknown_no_present_samples: bool = False,
                                    ignore_annotations: bool = False) -> pd.DataFrame:
        """
        Summarizes all mutations stored in this index relative to the passed reference genome.

        :param reference_name: The reference genome.
        :param id_type: The type of identifier to use.
        :param include_present_features: Whether or not mutation features present in this index (i.e., not unknown/missing)
                                should be included.
        :param include_unknown_features: Whether or not unknown mutations should be included.
        :param include_unknown_samples: Whether or not counts for those samples where it is unknown if they have a
                                        a feature should be included.
        :param include_unknown_no_present_samples: Whether or not counts for features where there are some unknowns but
                                                   no present samples should be included.
        :param ignore_annotations: Whether or not mutation annotations should be ignored.

        :return: A summary of all mutations in this index as a DataFrame.
        """
        features_summarizier = MutationFeaturesFromIndexComparator(connection=self._connection,
                                                                   include_unknown=include_unknown_features,
                                                                   include_present=include_present_features,
                                                                   include_unknown_samples=include_unknown_samples,
                                                                   include_unknown_no_present_samples=include_unknown_no_present_samples,
                                                                   id_type=id_type,
                                                                   ignore_annotations=ignore_annotations)
        if reference_name is None:
            sample_set = SampleSet.create_all()
        else:
            sample_set = self._connection.sample_service.get_samples_set_associated_with_reference(reference_name)

        return features_summarizier.summary(sample_set)

    def db_size(self, unit: str = 'B') -> pd.DataFrame:
        """
        Gets the size of this genomics index broken apart into separate sections.

        :param unit: The units to use for reporting the sizes.

        :return: A DataFrame summarizing the size of this genomics index.
        """
        return self._connection.db_size(unit)

    def samples_query(self, universe: str = 'all', **kwargs) -> SamplesQuery:
        """
        Constructs a new SamplesQuery with respect to this index.

        :param universe: The universe of samples under consideration. Options include ['all', 'mutations', 'dataframe'].
        :param kwargs: Additional arguments depending on the universe of samples.
        :return: A SamplesQuery object which can be used to further refine the query.
        """
        if universe == 'all':
            return self._query_all_samples(self._connection)
        elif universe == 'mutations':
            return self._query_reference(kind=universe, connection=self._connection, **kwargs)
        elif universe == 'mutations_experimental':
            return self._query_reference(kind=universe, connection=self._connection, **kwargs)
        elif universe == 'dataframe':
            return self._query_data_frame(connection=self._connection, **kwargs)
        else:
            raise Exception(f'Invalid universe=[{universe}]. Must be one of {self.QUERY_UNIVERSE}')

    def _query_all_samples(self, connection: DataIndexConnection):
        all_samples = connection.sample_service.get_all_sample_ids()
        return SamplesQueryIndex(connection=connection, sample_set=all_samples, universe_set=all_samples,
                                 unknown_set=SampleSet.create_empty())

    def _query_reference(self, kind: str, connection: DataIndexConnection, reference_name: str):
        reference_samples = connection.sample_service.get_samples_set_associated_with_reference(reference_name)
        reference_genome = connection.reference_service.find_reference_genome(reference_name)

        sample_query = SamplesQueryIndex(connection=connection, sample_set=reference_samples,
                                         unknown_set=SampleSet.create_empty(),
                                         universe_set=reference_samples)

        if reference_genome.has_tree():
            sample_query = TreeSamplesQueryFactory.instance().create_from_reference_genome(kind=kind,
                                                                                           reference_genome=reference_genome,
                                                                                           connection=connection,
                                                                                           wrapped_query=sample_query,
                                                                                           include_reference=True)

        return sample_query

    def _query_data_frame(self, connection: DataIndexConnection,
                          data_frame: pd.DataFrame = None,
                          sample_ids_column=None,
                          sample_names_column=None
                          ):
        if data_frame is None:
            raise Exception('data_frame must be set when querying with universe="dataframe"')
        if sample_ids_column is None and sample_names_column is None:
            raise Exception(
                'If querying with universe="dataframe", then one of sample_names_column or sample_ids_column '
                'must be set')
        elif sample_ids_column is not None:
            sample_query = self._query_all_samples(connection=connection)
            return DataFrameSamplesQuery.create_with_sample_ids_column(sample_ids_column,
                                                                       data_frame=data_frame,
                                                                       wrapped_query=sample_query,
                                                                       connection=connection)
        else:
            sample_query = self._query_all_samples(connection=connection)
            return DataFrameSamplesQuery.create_with_sample_names_column(sample_names_column,
                                                                         data_frame=data_frame,
                                                                         wrapped_query=sample_query,
                                                                         connection=connection)

    @classmethod
    def connect(cls, project_dir: Union[Path, str] = None, project: Project = None) -> GenomicsDataIndex:
        """
        Connects to a new genomics index. One of either project_dir or project must be set (by default assumes
        project_dir is the current directory).

        :param project_dir: The project/index directory. Defaults to the current directory.
        :param project: The :py:class:`genomics_data_index.configuration.Project` to connect to.
        :return: A new GenomicsDataIndex connected to the given project.
        """
        if project_dir is None and project is None:
            project_dir = os.getcwd()
            logger.warning(f'No project_dir or project specified. Assuming project is current dir [{project_dir}]')
            data_store_project = Project(project_dir)
        elif project_dir is not None and project is None:
            if isinstance(project_dir, str):
                project_dir = Path(project_dir)

            if not project_dir.exists():
                raise Exception(f'project_dir=[{project_dir}] does not exist')
            data_store_project = Project(root_dir=project_dir)
        else:
            data_store_project = project

        database_connection = data_store_project.create_connection()
        return GenomicsDataIndex(connection=database_connection)

    def __str__(self) -> str:
        samples_count = self._samples_count
        return f'<{self.__class__.__name__}(samples={samples_count})>'

    def __repr__(self) -> str:
        return str(self)
